# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Variant utilities."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import itertools

import enum
import six

from nucleus.protos import variants_pb2
from nucleus.util import ranges
from nucleus.util import variantcall_utils
from nucleus.util import vcf_constants


def only_call(variant):
  """Ensures the Variant has exactly one VariantCall, and returns it.

  Args:
    variant: nucleus.genomics.v1.Variant. The variant of interest.

  Returns:
    The single nucleus.genomics.v1.VariantCall in the variant.

  Raises:
    ValueError: Not exactly one VariantCall is in the variant.
  """
  if len(variant.calls) != 1:
    raise ValueError('Expected exactly one VariantCall in {}'.format(variant))
  return variant.calls[0]


def decode_variants(encoded_iter):
  """Yields a genomics.Variant from encoded_iter.

  Args:
    encoded_iter: An iterable that produces binary encoded
      nucleus.genomics.v1.Variant strings.

  Yields:
    A parsed nucleus.genomics.v1.Variant for each encoded element of
    encoded_iter in order.
  """
  for encoded in encoded_iter:
    yield variants_pb2.Variant.FromString(encoded)


def variant_position(variant):
  """Returns a new Range at the start position of variant.

  Args:
    variant: nucleus.genomics.v1.Variant.

  Returns:
    A new Range with the same reference_name as variant and start but an end
    that is start + 1. This produces a range that is the single basepair of the
    start of variant, hence the name position.
  """
  return ranges.make_range(variant.reference_name, variant.start,
                           variant.start + 1)


def variant_range(variant):
  """Returns a new Range covering variant.

  Args:
    variant: nucleus.genomics.v1.Variant.

  Returns:
    A new Range with the same reference_name, start, and end as variant.
  """
  return ranges.make_range(variant.reference_name, variant.start, variant.end)


def variant_range_tuple(variant):
  """Returns a new tuple of (reference_name, start, end) for the variant.

  A common use case for this function is to sort variants by chromosomal
  location, with usage like `sorted(variants, key=variant_range_tuple)`.

  Args:
    variant: nucleus.genomics.v1.Variant.

  Returns:
    A three-tuple with the same reference_name, start, and end as variant.
  """
  return (variant.reference_name, variant.start, variant.end)


@enum.unique
class GenotypeType(enum.Enum):
  """An enumeration of the types of genotypes."""
  hom_ref = ('homozygous reference', [0, 0], 0)
  het = ('heterozygous', [0, 1], 1)
  hom_var = ('homozygous non-reference', [1, 1], 2)
  no_call = ('no call', [-1, -1], -1)

  def __init__(self, full_name, example_gt, class_id):
    """Create a GenotypeType with the given name, GT and class_id."""
    self.full_name = full_name
    self.example_gt = example_gt
    self.class_id = class_id


@enum.unique
class VariantType(enum.Enum):
  """An enumeration of the types of variants."""
  # A variant.proto where there is no alt allele.
  ref = 0
  # A non-reference variant.proto where all ref and alt alleles are single
  # basepairs.
  snp = 1
  # A non-reference variant.proto where at least one of ref or alt alleles are
  # longer than 1 bp.
  indel = 2


def format_filters(variant):
  """Returns a human-readable string showing the filters applied to variant.

  Returns a string with the filter field values of variant separated by commas.
  If the filter field isn't set, returns vcf_constants.MISSING_FIELD ('.').

  Args:
    variant: nucleus.genomics.v1.Variant.

  Returns:
    A string.
  """
  if variant.filter:
    return ','.join(variant.filter)
  else:
    return vcf_constants.MISSING_FIELD


def format_alleles(variant):
  """Gets a string representation of the variant's alleles.

  Args:
    variant: nucleus.genomics.v1.Variant.

  Returns:
    A string ref_bases/alt1,alt2 etc.
  """
  return '{}/{}'.format(variant.reference_bases, ','.join(
      variant.alternate_bases))


def format_position(variant):
  """Gets a string representation of the variant's position.

  Args:
    variant: nucleus.genomics.v1.Variant.

  Returns:
    A string chr:start + 1 (as start is zero-based).
  """
  return '{}:{}'.format(variant.reference_name, variant.start + 1)


def _non_excluded_alts(alts, exclude_alleles=None):
  """Exclude any alts listed, by default: '<*>', '.', and '<NON_REF>'.

  These alleles are sometimes listed in ALT column but they shouldn't be
  analyzed and usually indicate reference blocks in formats like gVCF.

  E.g. 'A'->'<*>' is NOT an insertion, and 'A'->'.' is NOT a SNP.

  Args:
    alts: a list of strings representing the alternate alleles.
    exclude_alleles: list(str). The alleles in this list will be ignored.

  Returns:
    alts alleles except those in exclude_alleles, by default excluding the GVCF
    '<*>' allele, the '<NON_REF>' symbolic allele, and '.' missing field by
    default.
  """
  if exclude_alleles is None:
    exclude_alleles = [
        vcf_constants.GVCF_ALT_ALLELE, vcf_constants.SYMBOLIC_ALT_ALLELE,
        vcf_constants.MISSING_FIELD
    ]
  return [a for a in alts if a not in exclude_alleles]


def is_snp(variant, exclude_alleles=None):
  """Is variant a SNP?

  Args:
    variant: nucleus.genomics.v1.Variant.
    exclude_alleles: list(str). The alleles in this list will be ignored.

  Returns:
    True if all alleles of variant are 1 bp in length.
  """
  relevant_alts = _non_excluded_alts(variant.alternate_bases, exclude_alleles)
  return (len(variant.reference_bases) == 1 and len(relevant_alts) >= 1 and
          all(len(x) == 1 for x in relevant_alts))


def is_indel(variant, exclude_alleles=None):
  """Is variant an indel?

  An indel event is simply one where the size of at least one of the alleles
  is > 1.

  Args:
    variant: nucleus.genomics.v1.Variant.
    exclude_alleles: list(str). The alleles in this list will be ignored.

  Returns:
    True if the alleles in variant indicate an insertion/deletion event
    occurs at this site.
  """
  relevant_alts = _non_excluded_alts(variant.alternate_bases, exclude_alleles)
  if not relevant_alts:
    return False
  return (len(variant.reference_bases) > 1 or
          any(len(alt) > 1 for alt in relevant_alts))


def is_biallelic(variant, exclude_alleles=None):
  """Returns True if variant has exactly one alternate allele.

  Args:
    variant: nucleus.genomics.v1.Variant.
    exclude_alleles: list(str). The alleles in this list will be ignored.

  Returns:
    True if the variant has exactly one alternate allele.
  """
  relevant_alts = _non_excluded_alts(variant.alternate_bases, exclude_alleles)
  return len(relevant_alts) == 1


def is_multiallelic(variant, exclude_alleles=None):
  """Does variant have multiple alt alleles?

  Args:
    variant: nucleus.genomics.v1.Variant.
    exclude_alleles: list(str). The alleles in this list will be ignored.

  Returns:
    True if variant has more than one alt allele.
  """
  relevant_alts = _non_excluded_alts(variant.alternate_bases, exclude_alleles)
  return len(relevant_alts) > 1


def variant_is_insertion(variant, exclude_alleles=None):
  """Are all the variant's alt alleles insertions?

  Args:
    variant: nucleus.genomics.v1.Variant.
    exclude_alleles: list(str). The alleles in this list will be ignored.

  Returns:
    True if variant has at least one alt allele and all alts are insertions.
  """
  relevant_alts = _non_excluded_alts(variant.alternate_bases, exclude_alleles)
  if not relevant_alts:
    return False
  return all(
      is_insertion(variant.reference_bases, alt) for alt in relevant_alts)


def variant_is_deletion(variant, exclude_alleles=None):
  """Are all the variant's alt alleles deletions?

  Args:
    variant: nucleus.genomics.v1.Variant.
    exclude_alleles: list(str). The alleles in this list will be ignored.

  Returns:
    True if variant has at least one alt allele and all alts are deletions.
  """
  relevant_alts = _non_excluded_alts(variant.alternate_bases, exclude_alleles)
  if not relevant_alts:
    return False
  return all(is_deletion(variant.reference_bases, alt) for alt in relevant_alts)


def is_ref(variant, exclude_alleles=None):
  """Returns true if variant is a reference record.

  Variant protos can encode sites that aren't actually mutations in the
  sample. For example, the record ref='A', alt='.' indicates that there is
  no mutation present (i.e., alt is the missing value).

  Args:
    variant: nucleus.genomics.v1.Variant.
    exclude_alleles: list(str). The alleles in this list will be ignored.

  Returns:
    True if there are no actual alternate alleles.
  """
  relevant_alts = _non_excluded_alts(variant.alternate_bases, exclude_alleles)
  return not relevant_alts


def variant_type(variant):
  """Gets the VariantType of variant.

  Args:
    variant: nucleus.genomics.v1.Variant.

  Returns:
    VariantType indicating the type of this variant.
  """
  if is_ref(variant):
    return VariantType.ref
  elif is_snp(variant):
    return VariantType.snp
  else:
    return VariantType.indel


def is_transition(allele1, allele2):
  """Is the pair of single bp alleles a transition?

  Args:
    allele1: A string of the first allele, must be 1 bp in length.
    allele2: A string of the second allele, must be 1 bp in length.

  Returns:
    True if allele1/allele2 are a transition SNP.

  Raises:
    ValueError: if allele1 and allele2 are equal or aren't 1 bp in length.
  """
  if allele1 == allele2:
    raise ValueError('Alleles must be unique:', allele1, allele2)
  if len(allele1) != 1:
    raise ValueError('Alleles must be 1 bp in length.', allele1)
  if len(allele2) != 1:
    raise ValueError('Alleles must be 1 bp in length.', allele2)

  alleles_set = {allele1, allele2}
  return any(alleles_set == x for x in [{'A', 'G'}, {'C', 'T'}])


def is_insertion(ref, alt):
  """Is alt an insertion w.r.t. ref?

  Args:
    ref: A string of the reference allele.
    alt: A string of the alternative allele.

  Returns:
    True if alt is an insertion w.r.t. ref.
  """
  return len(ref) < len(alt)


def is_deletion(ref, alt):
  """Is alt a deletion w.r.t. ref?

  Args:
    ref: A string of the reference allele.
    alt: A string of the alternative allele.

  Returns:
    True if alt is a deletion w.r.t. ref.
  """
  return len(ref) > len(alt)


def has_insertion(variant):
  """Does variant have an insertion?

  Args:
    variant: nucleus.genomics.v1.Variant.

  Returns:
    True if the alleles in variant indicate an insertion event
    occurs at this site.
  """
  ref = variant.reference_bases
  return (is_indel(variant) and
          any(is_insertion(ref, alt) for alt in variant.alternate_bases))


def has_deletion(variant):
  """Does variant have a deletion?

  Args:
    variant: nucleus.genomics.v1.Variant.

  Returns:
    True if the alleles in variant indicate an deletion event
    occurs at this site.
  """
  ref = variant.reference_bases
  return (is_indel(variant) and
          any(is_deletion(ref, alt) for alt in variant.alternate_bases))


@enum.unique
class AlleleMismatchType(enum.Enum):
  """An enumeration of the types of allele mismatches we detect."""
  # Duplicate alleles.
  duplicate_eval_alleles = 1
  duplicate_true_alleles = 2
  # Truth has an allele that doesn't match any allele in eval.
  unmatched_true_alleles = 3
  # Eval has an allele that doesn't match any allele in truth.
  unmatched_eval_alleles = 4


def allele_mismatches(evalv, truev):
  """Determines the set of allele mismatch discordances between evalv and truev.

  Compares the alleles present in evalv and truev to determine if there are any
  disagreements between the set of called alleles in the two Variant protos. The
  type of differences basically boil down to:

  -- Are there duplicate alt alleles?
  -- Can we find a matching allele in the truev for each allele in evalv, and
    vice versa?

  Two alleles A and B match when they would produce the same sequence of bases
  in ref and alt haplotypes starting at the same position. So CA=>TA is the same
  as C=>T (position is the same, replacing A by A is a noop) but AC=>AT isn't
  the same as C=>T because the former event changes bases 1 bp further along in
  the reference genome than the C=>T allele.

  Args:
    evalv: A nucleus.genomics.v1.Variant.
    truev: A nucleus.genomics.v1.Variant.

  Returns:
    A set of AlleleMismatchType values.
  """
  unmatched_eval_alleles = []
  # Use set removes duplicate alleles in truth and eval variants.
  allele_matches = {alt: [] for alt in set(truev.alternate_bases)}
  for eval_alt in set(evalv.alternate_bases):
    # Loop over each possible alt allele, adding eval_alt to each matching alt
    # allele.
    found_match = False
    for true_alt in allele_matches:
      if (simplify_alleles(evalv.reference_bases, eval_alt) == simplify_alleles(
          truev.reference_bases, true_alt)):
        # We are a match to true_alt, so record that fact in allele_matches
        allele_matches[true_alt].append(eval_alt)
        found_match = True
    if not found_match:
      # We never found a match for eval_alt.
      unmatched_eval_alleles.append(eval_alt)

  # At this point we've checked every alt against every eval allele, and are
  # ready to summarize the differences using our AlleleMismatchType enum.
  types = set()
  if len(set(evalv.alternate_bases)) != len(evalv.alternate_bases):
    types.add(AlleleMismatchType.duplicate_eval_alleles)
  if len(set(truev.alternate_bases)) != len(truev.alternate_bases):
    types.add(AlleleMismatchType.duplicate_true_alleles)
  if unmatched_eval_alleles:
    types.add(AlleleMismatchType.unmatched_eval_alleles)
  if any(len(match) != 1 for match in six.itervalues(allele_matches)):
    types.add(AlleleMismatchType.unmatched_true_alleles)
  return types


def simplify_alleles(*alleles):
  """Simplifies alleles by stripping off common postfix bases.

  For example, simplify("AC", "GC") would produce the tuple "A", "G" as the "C"
  base is a common postfix of both alleles. But simplify("AC", "GT") would
  produce "AC", "GT" as there is no common postfix.

  Note this function will never simplify any allele down to the empty string. So
  if alleles = ['CACA', 'CA'], the longest common postfix is 'CA' but we will
  not produce ['CA', ''] as this is an invalid Variant allele encoding. Instead
  we produce ['CAC', 'C'].

  Args:
    *alleles: A tuple of bases, each as a string, to simplify.

  Returns:
    A tuple, one for each allele in alleles in order, with any common postfix
    bases stripped off.
  """

  def all_the_same(items):
    first = next(items)
    return all(item == first for item in items)

  # Loop over the alleles to determine the length of the shared postfix. Start
  # at 1 so every allele, even after trimming the postfix, has at least len 1.
  # For example, alleles = ['ATT', 'TT'] reduces to ['AT', 'T'] not ['A', ''].
  shortest_allele_len = min(len(a) for a in alleles)
  common_postfix_len = 0
  for i in range(1, shortest_allele_len):
    if not all_the_same(a[-i] for a in alleles):
      break
    common_postfix_len = i

  if common_postfix_len:
    return tuple(a[0:-common_postfix_len] for a in alleles)
  else:
    # Fast path for the case where there's no shared postfix.
    return alleles


def simplify_variant_alleles(variant):
  """Replaces the alleles in variants with their simplified versions.

  This function takes a variant and replaces its ref and alt alleles with those
  produced by a call to variant_utils.simplify_alleles() to remove common
  postfix bases in the alleles that may be present due to pruning away alleles.

  Args:
    variant: learning.genomics.genomics.Variant proto we want to simplify.

  Returns:
    variant with its ref and alt alleles replaced with their simplified
      equivalents.
  """
  simplified_alleles = simplify_alleles(variant.reference_bases,
                                        *variant.alternate_bases)
  variant.reference_bases = simplified_alleles[0]
  variant.alternate_bases[:] = simplified_alleles[1:]
  variant.end = variant.start + len(variant.reference_bases)
  return variant


def is_filtered(variant):
  """Returns True if variant has a non-PASS filter field, or False otherwise."""
  return bool(variant.filter) and any(
      f not in {'PASS', vcf_constants.MISSING_FIELD} for f in variant.filter)


def is_variant_call(variant,
                    require_non_ref_genotype=True,
                    no_calls_are_variant=False,
                    call_indices=None,
                    apply_filter=True):
  """Is variant a non-reference call?

  A Variant proto doesn't always imply that there's a variant present in the
  genome. The call may not have alternate bases, may be filtered, may a have
  hom-ref genotype, etc. This function looks for all of those configurations
  and returns true iff the variant is asserting that a mutation is present
  in the same.

  Note that this code allows a variant without a calls field to be variant,
  but one with a genotype call must have a non-reference genotype to be
  considered variant (if require_non_ref_genotype is True, the default). If
  False, a variant that passes all of the site-level requirements for being
  a variant_call will return a True value, regardless of the genotypes, which
  means that we'll consider a site with a sample with a hom-ref or no-call site
  a variant call.

  Args:
    variant: nucleus.genomics.v1.Variant.
    require_non_ref_genotype: Should we require a site with a genotype call to
      have a non-reference (het, hom-var) genotype for the site to be considered
      a variant call?
    no_calls_are_variant: If a site has genotypes, should we consider no_call
      genotypes as being variant or not? e.g. -1/1 listed as ./. in VCF
    call_indices: A list of 0-based indices. If specified, only the calls
      at the given indices will be considered. The function will return
      True if any of those calls are variant.
    apply_filter: If set to True, will never treat this site as variant when
      any filter other than PASS or . is set.

  Returns:
    True if variant is really a mutation call.
  """
  if is_ref(variant):
    # No actual alt allele listed in ALT column
    return False
  elif apply_filter and is_filtered(variant):
    # Anything other than PASS or . in FILTER column
    return False
  elif not variant.calls or not require_non_ref_genotype:
    return True
  # All tests after this point should only look at genotype-based fields, as
  # we may have aborted out in the prev. line due to require_non_ref_genotype.
  else:
    # Check for non-ref genotypes and optionally no-call (-1) genotypes
    if call_indices is None:
      call_indices = range(len(variant.calls))
    for i in call_indices:
      for g in variant.calls[i].genotype:
        if g > 0 or (no_calls_are_variant and g < 0):
          return True
    return False


def has_calls(variant):
  """Does variant have any genotype calls?

  Args:
    variant: nucleus.genomics.v1.Variant.

  Returns:
    True if variant has one or more VariantCalls.
  """
  return bool(variant.calls)


def genotype_type(variant):
  """Gets the GenotypeType for variant.

  If variant doesn't have genotypes, returns no_call. Otherwise
  returns one of no_call, hom_ref, het, or hom_var depending on the
  status of the genotypes in the call field of variant.

  Args:
    variant: nucleus.genomics.v1.Variant.

  Returns:
    A GenotypeType.

  Raises:
    ValueError: If variant has more than one call (i.e., is multi-sample).
  """
  if not has_calls(variant):
    return GenotypeType.no_call
  elif len(variant.calls) > 1:
    raise ValueError('Unsupported: multiple genotypes found at', variant)
  else:
    gt = set(only_call(variant).genotype)
    if gt == {-1}:
      return GenotypeType.no_call
    elif gt == {0}:
      return GenotypeType.hom_ref
    elif len(gt) > 1:
      return GenotypeType.het
    else:
      return GenotypeType.hom_var


def genotype_as_alleles(variant, call_ix=0):
  """Gets genotype of the sample in variant as a list of actual alleles.

  Returns the alleles specified by the genotype indices of variant.calls[0].
  For example, if variant.reference_bases = 'A' and variant.alternative_bases
  = ['C'] and the genotypes are [0, 1], this function will return
  ['A', 'C'].

  Args:
    variant: nucleus.genomics.v1.Variant.
    call_ix: int. The index into the calls attribute indicating which
      VariantCall to return alleles for.

  Returns:
    A list of allele (string) from variant, one for each genotype in
    variant.calls[call_ix], in order.

  Raises:
    ValueError: If variant doesn't have a call at the specified index.
  """
  if not 0 <= call_ix < len(variant.calls):
    raise ValueError(
        'Unsupported: requesting call {} in variant with {} calls: {}'.format(
            call_ix, len(variant.calls), variant))
  else:
    # Genotypes are encoded as integers, where 0 is the reference allele,
    # indices > 0 refer to alt alleles, and the no-call genotypes is encoded
    # as -1 in the genotypes. This code relies on this encoding to quickly
    # reference into the alleles by adding 1 to the genotype index.
    alleles = ([vcf_constants.MISSING_FIELD, variant.reference_bases] +
               list(variant.alternate_bases))
    return [alleles[i + 1] for i in variant.calls[call_ix].genotype]


def unphase_all_genotypes(variant):
  """Sorts genotype and removes phasing bit of all calls in variant.

  This mutation is done in place rather than returning a different copy.

  Args:
    variant: nucleus.genomics.v1.Variant.

  Returns:
    The variant with unphased calls.
  """
  for c in variant.calls:
    c.is_phased = False
    c.genotype.sort()
  return variant


def is_gvcf(variant):
  """Returns true if variant encodes a standard gVCF reference block.

  This means in practice that variant has a single alternate allele that is the
  canonical gVCF allele vcf_constants.GVCF_ALT_ALLELE.

  Args:
    variant: nucleus.genomics.v1.Variant.

  Returns:
    Boolean. True if variant is a gVCF record, False otherwise.
  """
  return variant.alternate_bases == [vcf_constants.GVCF_ALT_ALLELE]


def _genotype_order_in_likelihoods(num_alts, ploidy=2):
  """Yields tuples of `ploidy` ints for the given number of alt alleles.

  https://samtools.github.io/hts-specs/VCFv4.1.pdf
  "If A is the allele in REF and B,C,... are the alleles as ordered in ALT,
  the ordering of genotypes for the likelihoods is given by:
  F(j/k) = (k*(k+1)/2)+j. In other words, for biallelic sites the ordering is:
  AA,AB,BB; for triallelic sites the ordering is: AA,AB,BB,AC,BC,CC, etc."
  The biallelic sites in our case are 0/0, 0/1, 1/1.
  The triallelic sites are 0/0, 0/1, 1/1, 0/2, 1/2, 2/2.
  This wiki page has more information that generalizes to different ploidy.
  http://genome.sph.umich.edu/wiki/Relationship_between_Ploidy,_Alleles_and_Genotypes

  Args:
    num_alts: int. The number of alternate alleles at the site.
    ploidy: int. The ploidy for which to return genotypes.

  Yields:
    Tuples of `ploidy` ints representing allele indices in the order they appear
    in the corresponding genotype likelihood array.
  """
  if ploidy == 1:
    for i in range(num_alts + 1):
      yield (i,)
  elif ploidy == 2:
    for j in range(num_alts + 1):
      for i in range(j + 1):
        yield (i, j)
  else:
    raise NotImplementedError('Only haploid and diploid supported.')


def genotype_ordering_in_likelihoods(variant):
  """Yields (i, j, allele_i, allele_j) for the genotypes ordering in GLs.

  https://samtools.github.io/hts-specs/VCFv4.1.pdf
  "If A is the allele in REF and B,C,... are the alleles as ordered in ALT,
  the ordering of genotypes for the likelihoods is given by:
  F(j/k) = (k*(k+1)/2)+j. In other words, for biallelic sites the ordering is:
  AA,AB,BB; for triallelic sites the ordering is: AA,AB,BB,AC,BC,CC, etc."
  The biallelic sites in our case are 0/0, 0/1, 1/1.
  The triallelic sites are 0/0, 0/1, 1/1, 0/2, 1/2, 2/2.
  This wiki page has more information that generalizes ot different ploidy.
  http://genome.sph.umich.edu/wiki/Relationship_between_Ploidy,_Alleles_and_Genotypes

  Currently this function only implements for diploid cases.

  Args:
    variant: nucleus.genomics.v1.Variant.

  Yields:
    allele indices and strings (i, j, allele_i, allele_j) in the correct order.
  """
  alleles = [variant.reference_bases] + list(variant.alternate_bases)
  for i, j in _genotype_order_in_likelihoods(
      len(variant.alternate_bases), ploidy=2):
    yield i, j, alleles[i], alleles[j]


def genotype_likelihood(variant_call, allele_indices):
  """Returns the genotype likelihood for the given allele indices.

  Args:
    variant_call: nucleus.genomics.v1.VariantCall. The VariantCall from
      which to extract the genotype likelihood of the allele indices.
    allele_indices: list(int). The list of allele indices for a given genotype.
      E.g. diploid heterozygous alternate can be represented as [0, 1].

  Returns:
    The float value of the genotype likelihood of this set of alleles.
  """
  return variant_call.genotype_likelihood[genotype_likelihood_index(
      allele_indices)]


def genotype_likelihood_index(allele_indices):
  """Returns the genotype likelihood index for the given allele indices.

  Args:
    allele_indices: list(int). The list of allele indices for a given genotype.
      E.g. diploid homozygous reference is represented as [0, 0].

  Returns:
    The index into the associated genotype likelihood array corresponding to
    the likelihood of this list of alleles.

  Raises:
    NotImplementedError: The allele_indices are more than diploid.
  """
  if len(allele_indices) == 1:
    # Haploid case.
    return allele_indices[0]
  elif len(allele_indices) == 2:
    # Diploid case.
    g1, g2 = sorted(allele_indices)
    return g1 + (g2 * (g2 + 1) // 2)
  else:
    raise NotImplementedError(
        'Genotype likelihood index only supports haploid and diploid: {}'.
        format(allele_indices))


def allele_indices_for_genotype_likelihood_index(gl_index, ploidy=2):
  """Returns a tuple of allele_indices corresponding to the given GL index.

  This is the inverse function to `genotype_likelihood_index`.

  Args:
    gl_index: int. The index within a genotype likelihood array for which to
      determine the associated alleles.
    ploidy: int. The ploidy of the result.

  Returns:
    A tuple of `ploidy` ints representing the allele indices at this GL index.

  Raises:
    NotImplementedError: The requested allele indices are more than diploid.
  """
  if ploidy == 1:
    return gl_index
  elif ploidy == 2:
    # TODO(cym): Implement using algorithm described at
    # https://genome.sph.umich.edu/wiki/Relationship_between_Ploidy,_Alleles_and_Genotypes
    # rather than creating all genotypes explicitly.
    num_alts = 1
    while genotype_likelihood_index([num_alts, num_alts]) < gl_index:
      num_alts += 1
    genotypes = list(_genotype_order_in_likelihoods(num_alts, ploidy=ploidy))
    return genotypes[gl_index]
  else:
    raise NotImplementedError(
        'Allele calculations only supported for haploid and diploid.')


def allele_indices_with_num_alts(variant, num_alts, ploidy=2):
  """Returns a list of allele indices configurations with `num_alts` alternates.

  Args:
    variant: nucleus.genomics.v1.Variant. The variant of interest, which
      defines the candidate alternate alleles that can be used to generate
      allele indices configurations.
    num_alts: int in [0, `ploidy`]. The number of non-reference alleles for
      which to create the allele indices configurations.
    ploidy: int. The ploidy for which to return allele indices configurations.

  Returns: A list of tuples. Each tuple is of length `ploidy` and represents the
    allele indices of all `ploidy` genotypes that contain `num_alts`
    non-reference alleles.

  Raises:
    ValueError: The domain of `num_alts` is invalid.
    NotImplementedError: `ploidy` is not diploid.
  """
  if ploidy != 2:
    raise NotImplementedError(
        'allele_indices_with_num_alts only supports diploid.')
  if not 0 <= num_alts <= ploidy:
    raise ValueError(
        'Invalid number of alternate alleles requested: {} for ploidy {}'.
        format(num_alts, ploidy))

  max_candidate_alt_ix = len(variant.alternate_bases)
  if num_alts == 0:
    return [(0, 0)]
  elif num_alts == 1:
    return [(0, i) for i in range(1, max_candidate_alt_ix + 1)]
  else:
    return [(i, j)
            for i in range(1, max_candidate_alt_ix + 1)
            for j in range(i, max_candidate_alt_ix + 1)]


def variants_overlap(variant1, variant2):
  """Returns True if the range of variant1 and variant2 overlap.

  This is equivalent to:

    ranges_overlap(variant_range(variant1), variant_range(variant2))

  Args:
    variant1: nucleus.genomics.v1.Variant we want to compare for overlap.
    variant2: nucleus.genomics.v1.Variant we want to compare for overlap.

  Returns:
    True if the variants overlap, False otherwise.
  """
  return ranges.ranges_overlap(variant_range(variant1), variant_range(variant2))


def variant_key(variant, sort_alleles=True):
  """Gets a human-readable string key that is almost unique for Variant.

  Gets a string key that contains key information about the variant, formatted
  as:

    reference_name:start+1:reference_bases->alternative_bases

  where alternative bases is joined with a '/' for each entry in
  alternative_bases. The start+1 is so we display the position, which starts at
  1, and not the offset, which starts at 0.

  For example, a Variant(reference_name='20', start=10, reference_bases='AC',
  alternative_bases=['A', 'ACC']) would have a key of:

    20:11:AC->A/ACC

  The key is 'almost unique' in that the reference_name + start + alleles should
  generally occur once within a single VCF file, given the way the VCF
  specification works.

  Args:
    variant: nucleus.genomics.v1.Variant to make into a key.
    sort_alleles: bool. If True, the alternative_bases of variant will be sorted
      according to their lexicographic order. If False, the alternative_bases
      will be displayed in their order in the Variant.

  Returns:
    A str.
  """
  alts = variant.alternate_bases
  if sort_alleles:
    alts = sorted(alts)
  return '{}:{}:{}->{}'.format(variant.reference_name, variant.start + 1,
                               variant.reference_bases, '/'.join(alts))


def sorted_variants(variants):
  """Returns sorted(variants, key=variant_range_tuple)."""
  return sorted(variants, key=variant_range_tuple)


def variants_are_sorted(variants):
  """Returns True if variants are sorted w.r.t. variant_range.

  Args:
    variants: list[nucleus.genomics.v1.Variant]. A list of Variant
      protos that may or may not be sorted.

  Returns:
    True if variants are sorted, False otherwise.
  """
  def _pairwise(iterable):
    a, b = itertools.tee(iterable)
    next(b, None)
    return six.moves.zip(a, b)

  for r1, r2 in _pairwise(variant_range_tuple(v) for v in variants):
    if r2 < r1:
      return False
  return True


def set_info(variant, field_name, value, vcf_object=None):
  """Sets a field of the info map of the `Variant` to the given value(s).

  `variant.info` is analogous to the INFO field of a VCF record.

  Args:
    variant: Variant proto. The Variant to modify.
    field_name: str. The name of the field to set.
    value: A single value or list of values to update the Variant with. The type
      of the value is determined by the `vcf_object` if one is given, otherwise
      is looked up based on the reserved INFO fields in the VCF specification.
    vcf_object: (Optional) A VcfReader or VcfWriter object. If not None, the
      type of the field is inferred from the associated VcfReader or VcfWriter
      based on its name. Otherwise, the type is inferred if it is a reserved
      field.
  """
  if vcf_object is None:
    set_field_fn = vcf_constants.reserved_info_field_set_fn(field_name)
  else:
    set_field_fn = vcf_object.field_access_cache.info_field_set_fn(field_name)
  set_field_fn(variant.info, field_name, value)


def get_info(variant, field_name, vcf_object=None):
  """Returns the value of the `field_name` INFO field.

  The `vcf_object` is used to determine the type of the resulting value. If it
  is a single value or a Flag, that single value will be returned. Otherwise,
  the list of values is returned.

  Args:
    variant: Variant proto. The Variant of interest.
    field_name: str. The name of the field to retrieve values from.
    vcf_object: (Optional) A VcfReader or VcfWriter object. If not None, the
      type of the field is inferred from the associated VcfReader or VcfWriter
      based on its name. Otherwise, the type is inferred if it is a reserved
      field.
  """
  if vcf_object is None:
    get_field_fn = vcf_constants.reserved_info_field_get_fn(field_name)
  else:
    get_field_fn = vcf_object.field_access_cache.info_field_get_fn(field_name)
  return get_field_fn(variant.info, field_name)


def calc_ac(variant):
  """Returns a list of alt counts based on variant.calls."""
  counts = [0] * len(variant.alternate_bases)
  for call in variant.calls:
    for gt in call.genotype:
      if gt > 0:
        counts[gt - 1] += 1
  return counts


def calc_an(variant):
  """Returns the total number of alleles in called genotypes in variant."""
  return sum(
      len([1 for gt in call.genotype if gt > -1]) for call in variant.calls)


def is_singleton(variant):
  """Returns True iff the variant has exactly one non-ref VariantCall."""
  non_ref_count = 0
  for c in variant.calls:
    if variantcall_utils.has_variation(c):
      non_ref_count += 1
      if non_ref_count > 1:
        return False
  return non_ref_count == 1


def major_allele_frequency(variant):
  """Returns the frequency of the most common allele in the variant."""
  count_dict = collections.defaultdict(int)
  for call in variant.calls:
    for geno in call.genotype:
      count_dict[geno] += 1
  denom = sum(cnt for geno, cnt in count_dict.items() if geno >= 0)
  if denom > 0:
    numer = max(cnt for geno, cnt in count_dict.items() if geno >= 0)
    return float(numer) / denom
  else:
    return 0



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes for reading and writing BED files.

The BED format is described at
https://genome.ucsc.edu/FAQ/FAQformat.html#format1

API for reading:

```python
from nucleus.io import bed

# Iterate through all records.
with bed.BedReader(input_path) as reader:
  for record in reader:
    print(record)
```

where `record` is a `nucleus.genomics.v1.BedRecord` protocol buffer.

API for writing:

```python
from nucleus.io import bed
from nucleus.protos import bed_pb2

# records is an iterable of nucleus.genomics.v1.BedRecord protocol buffers.
records = ...

# header defines how many fields to write out.
header = bed_pb2.BedHeader(num_fields=5)

# Write all records to the desired output path.
with bed.BedWriter(output_path, header) as writer:
  for record in records:
    writer.write(record)
```

For both reading and writing, if the path provided to the constructor contains
'.tfrecord' as an extension, a `TFRecord` file is assumed and attempted to be
read or written. Otherwise, the filename is treated as a true BED file.

Files that end in a '.gz' suffix cause the file to be treated as compressed
(with BGZF if it is a true BED file, and with gzip if it is a TFRecord file).
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from nucleus.io import genomics_reader
from nucleus.io import genomics_writer
from nucleus.io.python import bed_reader
from nucleus.io.python import bed_writer
from nucleus.protos import bed_pb2


class NativeBedReader(genomics_reader.GenomicsReader):
  """Class for reading from native BED files.

  Most users will want to use BedReader instead, because it dynamically
  dispatches between reading native BED files and TFRecord files based on the
  filename's extension.
  """

  def __init__(self, input_path, num_fields=0):
    """Initializes a NativeBedReader.

    Args:
      input_path: string. A path to a resource containing BED records.
      num_fields: int. The number of fields to read in the BED. If unset or set
        to zero, all fields in the input are read.
    """
    super(NativeBedReader, self).__init__()

    bed_path = input_path.encode('utf8')
    options = bed_pb2.BedReaderOptions(num_fields=num_fields)
    self._reader = bed_reader.BedReader.from_file(bed_path, options)
    self.header = self._reader.header

  def query(self):
    """Returns an iterator for going through the records in the region.

    NOTE: This function is not currently implemented by NativeBedReader though
    it could be implemented for sorted, tabix-indexed BED files.
    """
    raise NotImplementedError('Can not currently query a BED file')

  def iterate(self):
    """Returns an iterable of BedRecord protos in the file."""
    return self._reader.iterate()

  def __exit__(self, exit_type, exit_value, exit_traceback):
    self._reader.__exit__(exit_type, exit_value, exit_traceback)


class BedReader(genomics_reader.DispatchingGenomicsReader):
  """Class for reading BedRecord protos from BED or TFRecord files."""

  def _native_reader(self, input_path, **kwargs):
    return NativeBedReader(input_path, **kwargs)

  def _record_proto(self):
    return bed_pb2.BedRecord


class NativeBedWriter(genomics_writer.GenomicsWriter):
  """Class for writing to native BED files.

  Most users will want BedWriter, which will write to either native BED
  files or TFRecord files, based on the output filename's extension.
  """

  def __init__(self, output_path, header=None):
    """Initializer for NativeBedWriter.

    Args:
      output_path: str. The path to which to write the BED file.
      header: nucleus.genomics.v1.BedHeader. The header that defines all
        information germane to the constituent BED records.
    """
    super(NativeBedWriter, self).__init__()
    if header is None:
      header = bed_pb2.BedHeader(num_fields=3)
    writer_options = bed_pb2.BedWriterOptions()
    self._writer = bed_writer.BedWriter.to_file(output_path, header,
                                                writer_options)

  def write(self, proto):
    self._writer.write(proto)

  def __exit__(self, exit_type, exit_value, exit_traceback):
    self._writer.__exit__(exit_type, exit_value, exit_traceback)


class BedWriter(genomics_writer.DispatchingGenomicsWriter):
  """Class for writing BedRecord protos to BED or TFRecord files."""

  def _native_writer(self, output_path, header):
    return NativeBedWriter(output_path, header=header)



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities to help with testing code."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

from absl import flags
from absl.testing import absltest
import six

from nucleus.io import gfile
from nucleus.protos import position_pb2
from nucleus.protos import reads_pb2
from nucleus.protos import struct_pb2
from nucleus.protos import variants_pb2
from nucleus.util import cigar as _cigar

FLAGS = flags.FLAGS

# In the OSS version these will be ''.
DATADIR = ''
DEFAULT_WORKSPACE = ''

# In the OSS version this becomes 'nucleus/testdata'
RELATIVE_TESTDATA_PATH = 'nucleus/testdata'


def genomics_testdata(path, datadir=DATADIR):
  """Gets the path to a testdata file in genomics at relative path.

  Args:
    path: A path to a testdata file *relative* to the genomics root
      directory. For example, if you have a test file in
      "datadir/nucleus/testdata/foo.txt", path should be
      "nucleus/testdata/foo.txt" to get a path to it.
    datadir: The path of the genomics root directory *relative* to
      the testing source directory.

  Returns:
    The absolute path to a testdata file.
  """
  if hasattr(FLAGS, 'test_srcdir'):
    # Google code uses FLAG.test_srcdir
    # TensorFlow uses a routine googletest.test_src_dir_path.
    test_workspace = os.environ.get('TEST_WORKSPACE', DEFAULT_WORKSPACE)
    test_srcdir = os.path.join(FLAGS.test_srcdir, test_workspace)
  else:
    # In bazel TEST_SRCDIR points at the runfiles directory, and
    # TEST_WORKSPACE names the workspace.  We need to append to the
    # path the name of the workspace in order to get to the root of our
    # source tree.
    test_workspace = os.environ['TEST_WORKSPACE']
    test_srcdir = os.path.join(os.environ['TEST_SRCDIR'], test_workspace)
  return os.path.join(test_srcdir, datadir, path)


# TODO(mdepristo): is this necessary?
def genomics_core_testdata(filename):
  """Gets the path to a testdata named filename in util/testdata.

  Args:
    filename: The name of a testdata file in the core genomics testdata
      directory. For example, if you have a test file in
      "third_party/nucleus/util/testdata/foo.txt", filename should be
      "foo.txt" to get a path to it.

  Returns:
    The absolute path to a testdata file.
  """
  return genomics_testdata(os.path.join(RELATIVE_TESTDATA_PATH, filename))


def test_tmpfile(name, contents=None):
  """Returns a path to a tempfile named name in the test_tmpdir.

  Args:
    name: str; the name of the file, should not contain any slashes.
    contents: bytes, or None. If not None, tmpfile's contents will be set to
      contents before returning the path.

  Returns:
    str path to a tmpfile with filename name in our test tmpfile directory.
  """
  path = os.path.join(absltest.get_default_test_tmpdir(), name)
  if contents is not None:
    with gfile.Open(path, 'wb') as fout:
      fout.write(contents)
  return path


def set_list_values(list_value, values):
  """Sets a ListValue to have the values in values."""

  def format_one(value):
    if isinstance(value, str):
      return struct_pb2.Value(string_value=value)
    elif isinstance(value, float):
      return struct_pb2.Value(number_value=value)
    elif isinstance(value, six.integer_types):
      return struct_pb2.Value(int_value=value)
    else:
      raise ValueError('Unsupported type ', value)

  del list_value.values[:]
  list_value.values.extend([format_one(value) for value in values])
  # list_value.values.extend(vals)


def make_variant(chrom='chr1',
                 start=10,
                 alleles=None,
                 end=None,
                 filters=None,
                 qual=None,
                 gt=None,
                 gq=None,
                 sample_name=None,
                 gls=None,
                 is_phased=None):
  """Creates a new Variant proto from args.

  Args:
    chrom: str. The reference_name for this variant.
    start: int. The starting position of this variant.
    alleles: list of str with at least one element. alleles[0] is the reference
      bases and alleles[1:] will be set to alternate_bases of variant. If None,
      defaults to ['A', 'C'].
    end: int or None. If not None, the variant's end will be set to this value.
      If None, will be set to the start + len(reference_bases).
    filters: str, list of str, or None. Sets the filters field of the variant to
      this value if not None. If filters is a string `value`, this is equivalent
      to an argument [`value`]. If None, no value will be assigned to the
      filters field.
    qual: int or None. The quality score for this variant. If None, no quality
      score will be written in the Variant.
    gt: A list of ints, or None. If present, creates a VariantCall in Variant
      with genotype field set to this value. The special 'DEFAULT' value, if
      provided, will set the genotype to [0, 1]. This is the default behavior.
    gq: int or None. If not None and gt is not None, we will add an this GQ
      value to our VariantCall.
    sample_name: str or None. If not None and gt is not None, sets the
      call_set_name of our VariantCall to this value.
    gls: array-list of float, or None. If not None and gt is not None, sets the
      genotype_likelihoods of our VariantCall to this value.
    is_phased: bool. Indicates whether a VariantCall should be phased.

  Returns:
    nucleus.genomics.v1.Variant proto.
  """
  return make_variant_multiple_calls(
      chrom=chrom,
      start=start,
      alleles=alleles,
      end=end,
      filters=filters,
      qual=qual,
      gts=None if gt is None else [gt],
      gqs=None if gq is None else [gq],
      sample_names=None if sample_name is None else [sample_name],
      glss=None if gls is None else [gls],
      is_phased=None if is_phased is None else [is_phased])


def make_variant_multiple_calls(chrom='chr1',
                                start=10,
                                alleles=None,
                                end=None,
                                filters=None,
                                qual=None,
                                gts=None,
                                gqs=None,
                                sample_names=None,
                                glss=None,
                                is_phased=None):
  """Creates a new Variant proto from args that contains multi-sample calls.

  Args:
    chrom: str. The reference_name for this variant.
    start: int. The starting position of this variant.
    alleles: list of str with at least one element. alleles[0] is the reference
      bases and alleles[1:] will be set to alternate_bases of variant. If None,
        defaults to ['A', 'C'].
    end: int or None. If not None, the variant's end will be set to this value.
      If None, will be set to the start + len(reference_bases).
    filters: str, list of str, or None. Sets the filters field of the variant to
      this value if not None. If filters is a string `value`, this is equivalent
      to an argument [`value`]. If None, no value will be assigned to the
      filters field.
    qual: int or None. The quality score for this variant. If None, no quality
      score will be written in the Variant.
    gts: A list of lists of ints. For each list in this list, creates a
      VariantCall in Variant with genotype field set to this value.
    gqs: A list of ints or None. Must match the gts arg if specified. Sets the
      GQ value of corresponding VariantCall.
    sample_names: A list of strs or None. Must match the gts arg if specified.
      Sets the call_set_name of the corresponding VariantCall.
    glss: A list of array-lists of float, or None. Must match the gts arg if
      specified. Sets the genotype_likelihoods of the corresponding VariantCall.
    is_phased: list of bools. Must match the gts arg if specified. Indicates
      whether the corresponding VariantCall should be phased.

  Returns:
    nucleus.genomics.v1.Variant proto.
  """
  if alleles is None:
    alleles = ['A', 'C']

  if not end:
    end = start + len(alleles[0])

  variant = variants_pb2.Variant(
      reference_name=chrom,
      start=start,
      end=end,
      reference_bases=alleles[0],
      alternate_bases=alleles[1:],
      quality=qual,
  )

  if filters is not None:
    if not isinstance(filters, (list, tuple)):
      filters = [filters]
    variant.filter[:] = filters

  if gts:
    for i in range(len(gts)):
      call = variant.calls.add(genotype=gts[i])

      if sample_names and sample_names[i] is not None:
        call.call_set_name = sample_names[i]

      if gqs and gqs[i] is not None:
        set_list_values(call.info['GQ'], [gqs[i]])

      if glss and glss[i] is not None:
        call.genotype_likelihood.extend(glss[i])

      if is_phased and is_phased[i] is not None:
        call.is_phased = is_phased[i]

  return variant


def make_read(bases,
              start,
              quals=None,
              cigar=None,
              mapq=50,
              chrom='chr1',
              name=None):
  """Makes a nucleus.genomics.v1.Read for testing."""
  if quals and len(bases) != len(quals):
    raise ValueError('Incompatable bases and quals', bases, quals)
  read = reads_pb2.Read(
      fragment_name=name if name else 'read_' + str(make_read.counter),
      proper_placement=True,
      read_number=1,
      number_reads=2,
      aligned_sequence=bases,
      aligned_quality=quals,
      alignment=reads_pb2.LinearAlignment(
          position=position_pb2.Position(reference_name=chrom, position=start),
          mapping_quality=mapq,
          cigar=_cigar.to_cigar_units(cigar) if cigar else []))
  make_read.counter += 1
  return read
make_read.counter = 0


def cc_iterable_len(cc_iterable):
  """Count the number of elements in an Iterable object.

  Args:
    cc_iterable: a CLIF-wrap of a subclass of the C++ Iterable class.

  Returns:
    integer count
  """
  count = 0
  while True:
    not_done, _ = cc_iterable.Next()
    if not not_done:
      break
    count += 1
  return count


def iterable_len(iterable):
  """Returns the length of a Python iterable, by advancing it."""
  return sum(1 for _ in iterable)


# TODO(b/63955799): remove and replace uses when bug is fixed in mock.
def assert_not_called_workaround(mock):
  """Asserts that a mock has not been called.

  There's a bug in mock where some of the assert functions on a mock are being
  dropped when that mock is created with an autospec:

    https://bugs.python.org/issue28380

  The mock 2.0.0 backport doesn't have the fix yet. The required patch is:

    https://bugs.python.org/file44991/fix_autospecced_mock_functions.patch

  but the current mock (checked 07/22/17) backport code is missing the fix:

    https://github.com/testing-cabal/mock/blob/master/mock/mock.py#L315

  This is an open issue on the mock github repo:

    https://github.com/testing-cabal/mock/issues/398

  And they claim that it'll be a few months (as of April 2017) before it is
  incorporated into the backport.

  Args:
    mock: The mock to assert hasn't been called.

  Raises:
    AssertionError: mock has been called.
  """
  if mock.call_count != 0:
    raise AssertionError("Expected no calls to '{}' but was called {} times"
                         .format(mock.name, mock.call_count))


# TODO(b/63955799): remove and replace uses when bug is fixed in mock.
def assert_called_once_workaround(mock):
  """Asserts that a mock has been called exactly once.

  See assert_not_called_workaround for the backstory on why this function
  exists.

  Args:
    mock: The mock that should have been called exactly once.

  Raises:
    AssertionError: mock wasn't called exactly once.
  """
  if mock.call_count != 1:
    raise AssertionError(
        "Expected exactly one call to '{}' but was called {} times".format(
            mock.name, mock.call_count))



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for manipulating DNA sequences."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


class Error(Exception):
  """Base error class."""


def _add_lowercase(d):
  """Returns a dictionary with the lowercase keys and values entered."""
  retval = d.copy()
  retval.update({k.lower(): v.lower() for k, v in d.items()})
  return retval


STRICT_DNA_COMPLEMENT_UPPER = {'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C'}
DNA_COMPLEMENT_UPPER = {'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C', 'N': 'N'}
IUPAC_DNA_COMPLEMENT_UPPER = {
    'A': 'T',
    'T': 'A',
    'C': 'G',
    'G': 'C',
    'R': 'Y',  # R is A/G
    'Y': 'R',  # Y is C/T
    'S': 'S',  # S is C/G
    'W': 'W',  # W is A/T
    'K': 'M',  # K is G/T
    'M': 'K',  # M is A/C
    'B': 'V',  # B is C/G/T
    'V': 'B',  # V is A/C/G
    'D': 'H',  # D is A/G/T
    'H': 'D',  # H is A/C/T
    'N': 'N',  # N is any base
}

IUPAC_TO_CANONICAL_BASES_UPPER = {
    'A': ['A'],
    'T': ['T'],
    'C': ['C'],
    'G': ['G'],
    'R': ['A', 'G'],
    'Y': ['C', 'T'],
    'S': ['C', 'G'],
    'W': ['A', 'T'],
    'K': ['G', 'T'],
    'M': ['A', 'C'],
    'B': ['C', 'G', 'T'],
    'V': ['A', 'C', 'G'],
    'D': ['A', 'G', 'T'],
    'H': ['A', 'C', 'T'],
    'N': ['A', 'C', 'G', 'T'],
}

STRICT_DNA_COMPLEMENT = _add_lowercase(STRICT_DNA_COMPLEMENT_UPPER)
DNA_COMPLEMENT = _add_lowercase(DNA_COMPLEMENT_UPPER)
IUPAC_DNA_COMPLEMENT = _add_lowercase(IUPAC_DNA_COMPLEMENT_UPPER)


STRICT_DNA_BASES_UPPER = frozenset(['A', 'C', 'G', 'T'])
STRICT_DNA_BASES = frozenset(['a', 'c', 'g', 't', 'A', 'C', 'G', 'T'])
DNA_BASES_UPPER = frozenset(['A', 'C', 'G', 'T', 'N'])
DNA_BASES = frozenset(['a', 'c', 'g', 't', 'n', 'A', 'C', 'G', 'T', 'N'])


def reverse_complement(sequence, complement_dict=None):
  """Returns the reverse complement of a DNA sequence.

  By default this will successfully reverse complement sequences comprised
  solely of A, C, G, and T letters. Other complement dictionaries can be
  passed in for more permissive matching.

  Args:
    sequence: str. The input sequence to reverse complement.
    complement_dict: dict[str, str]. The lookup dictionary holding the
      complement base pairs.

  Returns:
    The reverse complement DNA sequence.

  Raises:
    Error: The sequence contains letters not present in complement_dict.
  """
  if complement_dict is None:
    complement_dict = STRICT_DNA_COMPLEMENT_UPPER

  try:
    return ''.join(complement_dict[nt] for nt in reversed(sequence))
  except KeyError:
    raise Error('Unknown base in {}, cannot reverse complement using {}'.format(
        sequence, str(complement_dict)))



# Copyright 2019 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Utility functions for visualization and inspection of pileup examples.

Visualization and inspection utility functions enable showing image-like array
data including those used in DeepVariant.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from IPython import display
import numpy as np
from PIL import Image
from PIL import ImageDraw

from nucleus.io import gfile
from nucleus.protos import variants_pb2

DEEPVARIANT_CHANNEL_NAMES = [
    'read base', 'base quality', 'mapping quality', 'strand',
    'read supports variant', 'base differs from ref', 'alternate allele 1',
    'alternate allele 2'
]


def get_image_array_from_example(example):
  """Decode image/encoded and image/shape of an Example into a numpy array.

  Parse image/encoded and image/shape features from a tensorflow Example and
  decode the image into that shape.

  Args:
    example: a tensorflow Example containing features that include
      "image/encoded" and "image/shape"

  Returns:
    numpy array of dtype np.uint8.
  """
  features = example.features.feature
  img = features['image/encoded'].bytes_list.value[0]
  shape = features['image/shape'].int64_list.value[0:3]
  return np.frombuffer(img, np.uint8).reshape(shape)


def split_3d_array_into_channels(arr):
  """Split 3D array into a list of 2D arrays.

  e.g. given a numpy array of shape (100, 200, 6), return a list of 6 channels,
  each with shape (100, 200).

  Args:
    arr: a 3D numpy array.

  Returns:
    list of 2D numpy arrays.
  """
  return [arr[:, :, i] for i in range(arr.shape[-1])]


def channels_from_example(example):
  """Extract image from an Example and return the list of channels.

  Args:
    example: a tensorflow Example containing features that include
      "image/encoded" and "image/shape"

  Returns:
    list of 2D numpy arrays, one for each channel.
  """
  image = get_image_array_from_example(example)
  return split_3d_array_into_channels(image)


def convert_6_channels_to_rgb(channels):
  """Convert 6-channel image from DeepVariant to RGB for quick visualization.

  The 6 channels are: "read base", "base quality", "mapping quality", "strand",
  "supports variant", "base != reference".

  Args:
    channels: a list of 6 numpy arrays.

  Returns:
    3D numpy array of 3 colors (Red, green, blue).
  """
  base = channels[0]
  # qual is the minimum of base quality and mapping quality at each position
  # 254 is the max value for quality scores because the SAM specification has
  # 255 reserved for unavailable values.
  qual = np.minimum(channels[1], channels[2])
  strand = channels[3]
  # alpha is <supports variant> * <base != reference>
  alpha = np.multiply(channels[4] / 254.0, channels[5] / 254.0)
  return np.multiply(np.stack([base, qual, strand]),
                     alpha).astype(np.uint8).transpose([1, 2, 0])


def scale_colors_for_png(arr, vmin=0, vmax=255):
  """Scale an array to integers between 0 and 255 to prep it for a PNG image.

  Args:
    arr: numpy array. Input array made up of integers or floats.
    vmin: number. Minimum data value to map to 0. Values below this will be
      clamped to this value and therefore become 0.
    vmax: number. Maximum data value to map to 255. Values above this will be
      clamped to this value and therefore become 255.

  Returns:
    numpy array of dtype np.uint8 (integers between 0 and 255).
  """
  if vmax == 0 or vmax <= vmin:
    raise ValueError('vmin must be non-zero and higher than vmin.')

  # Careful not to modify the original array
  scaled = np.copy(arr)

  # Snap numbers in the array falling outside the range into the range,
  # otherwise they will produce artifacts due to byte overflow
  scaled[scaled > vmax] = vmax
  scaled[scaled < vmin] = vmin

  # Scale the input into the range of vmin to vmax
  if vmin != 0 or vmax != 255:
    scaled = ((scaled - vmin) / (vmax - vmin)) * 255
  return scaled.astype(np.uint8)


def _get_image_type_from_array(arr):
  """Find image type based on array dimensions.

  Raises error on invalid image dimensions.
  Args:
    arr: numpy array. Input array.

  Returns:
    str. "RGB" or "L", meant for PIL.Image.fromarray.
  """
  if len(arr.shape) == 3 and arr.shape[2] == 3:
    # 8-bit x 3 colors
    return 'RGB'
  elif len(arr.shape) == 2:
    # 8-bit, gray-scale
    return 'L'
  else:
    raise ValueError(
        'Input array must have either 2 dimensions or 3 dimensions where the '
        'third dimension has 3 channels. i.e. arr.shape is (x,y) or (x,y,3). '
        'Found shape {}.'.format(arr.shape))


def autoscale_colors_for_png(arr, vmin=None, vmax=None):
  """Adjust an array to prepare it for saving to an image.

  Re-scale numbers in the input array to go from 0 to 255 to adapt them for a
  PNG image.

  Args:
    arr: numpy array. Should be 2-dimensional or 3-dimensional where the third
      dimension has 3 channels.
    vmin: number (float or int). Minimum data value, which will correspond to
      black in greyscale or lack of each color in RGB images. Default None takes
      the minimum of the data from arr.
    vmax: number (float or int). Maximum data value, which will correspond to
      white in greyscale or full presence of each color in RGB images. Default
      None takes the max of the data from arr.

  Returns:
    (modified numpy array, image_mode)
  """
  image_mode = _get_image_type_from_array(arr)

  if vmin is None:
    vmin = np.min(arr)
  if vmax is None:
    vmax = np.max(arr)

  # In cases where all elements are the same, fix the vmax so that even though
  # the whole image will be black, the user can at least see the shape
  if vmin == vmax:
    vmax = vmin + 1

  scaled = scale_colors_for_png(arr, vmin=vmin, vmax=vmax)
  return scaled, image_mode


def add_header(img, labels, mark_midpoints=True, header_height=20):
  """Adds labels to the image, evenly distributed across the top.

  This is primarily useful for showing the names of channels.

  Args:
    img: A PIL Image.
    labels: list of strs. Labels for segments to write across the top.
    mark_midpoints: bool. Whether to add a small vertical line marking the
      center of each segment of the image.
    header_height: int. Height of the header in pixels.

  Returns:
    A new PIL Image, taller than the original img and annotated.
  """

  # Create a taller image to make space for a header at the top.
  new_height = header_height + img.size[1]
  new_width = img.size[0]

  if img.mode == 'RGB':
    placeholder_size = (new_height, new_width, 3)
  else:
    placeholder_size = (new_height, new_width)
  placeholder = np.ones(placeholder_size, dtype=np.uint8) * 255

  # Divide the image width into segments.
  segment_width = img.size[0] / len(labels)

  # Calculate midpoints for all segments.
  midpoints = [int(segment_width * (i + 0.5)) for i in range(len(labels))]

  if mark_midpoints:
    # For each label, add a small line to mark the middle.
    for x_position in midpoints:
      placeholder[header_height - 5:header_height, x_position] = 0
      # If image has an even width, it will need 2 pixels marked as the middle.
      if segment_width % 2 == 0:
        placeholder[header_height - 5:header_height, x_position + 1] = 0

  bigger_img = Image.fromarray(placeholder, mode=img.mode)
  # Place the original image inside the taller placeholder image.
  bigger_img.paste(img, (0, header_height))

  # Add a label for each segment.
  draw = ImageDraw.Draw(bigger_img)
  for i in range(len(labels)):
    text = labels[i]
    text_width = draw.textsize(text)[0]
    # xy refers to the left top corner of the text, so to center the text on
    # the midpoint, subtract half the text width from the midpoint position.
    x_position = int(midpoints[i] - text_width / 2)
    draw.text(xy=(x_position, 0), text=text, fill='black')
  return bigger_img


def save_to_png(arr,
                path=None,
                image_mode=None,
                show=True,
                labels=None,
                scale=None):
  """Make a PNG and show it from a numpy array of dtype=np.uint8.

  Args:
    arr: numpy array. Input array to save.
    path: str. File path at which to save the image. A .png prefix is added if
      the path does not already have one. Leave empty to save at /tmp/tmp.png,
      which is useful when only temporarily showing the image in a Colab
      notebook.
    image_mode: "RGB" or "L". Leave as default=None to choose based on image
      dimensions.
    show: bool. Whether to display the image using IPython (for notebooks).
    labels: list of str. Labels to show across the top of the image.
    scale: integer. Number of pixels wide and tall to show each cell in the
      array. This sizes up the image while keeping exactly the same number of
      pixels for every cell in the array, preserving resolution and preventing
      any interpolation or overlapping of pixels. Default None adapts to the
      size of the image to multiply it up until a limit of 500 pixels, a
      convenient size for use in notebooks. If saving to a file for automated
      processing, scale=1 is recommended to keep output files small and simple
      while still retaining all the information content.

  Returns:
    None. Saves an image at path and optionally shows it with IPython.display.
  """
  if image_mode is None:
    image_mode = _get_image_type_from_array(arr)

  img = Image.fromarray(arr, mode=image_mode)

  if labels is not None:
    img = add_header(img, labels)

  if scale is None:
    scale = max(1, int(500 / max(arr.shape)))

  if scale != 1:
    img = img.resize((img.size[0] * scale, img.size[1] * scale))

  # Saving to a temporary file is needed even when showing in a notebook
  if path is None:
    path = '/tmp/tmp.png'
  elif not path.endswith('.png'):
    # Only PNG is supported because JPEG files are unnecessarily 3 times larger.
    path = '{}.png'.format(path)
  with gfile.Open(path, 'wb') as fout:
    img.save(fout, format=path.split('.')[-1])

  # Show image (great for notebooks)
  if show:
    display.display(display.Image(path))


def array_to_png(arr,
                 path=None,
                 show=True,
                 vmin=None,
                 vmax=None,
                 scale=None,
                 labels=None):
  """Save an array as a PNG image with PIL and show it.

  Args:
    arr: numpy array. Should be 2-dimensional or 3-dimensional where the third
      dimension has 3 channels.
    path: str. Path for the image output. Default is /tmp/tmp.png for quickly
      showing the image in a notebook.
    show: bool. Whether to show the image using IPython utilities, only works in
      notebooks.
    vmin: number. Minimum data value, which will correspond to black in
      greyscale or lack of each color in RGB images. Default None takes the
      minimum of the data from arr.
    vmax: number. Maximum data value, which will correspond to white in
      greyscale or full presence of each color in RGB images. Default None takes
      the max of the data from arr.
    scale: integer. Number of pixels wide and tall to show each cell in the
      array. This sizes up the image while keeping exactly the same number of
      pixels for every cell in the array, preserving resolution and preventing
      any interpolation or overlapping of pixels. Default None adapts to the
      size of the image to multiply it up until a limit of 500 pixels, a
      convenient size for use in notebooks. If saving to a file for automated
      processing, scale=1 is recommended to keep output files small and simple
      while still retaining all the information content.
    labels: list of str. Labels to show across the top of the image.

  Returns:
    None. Saves an image at path and optionally shows it with IPython.display.
  """
  scaled, image_mode = autoscale_colors_for_png(arr, vmin=vmin, vmax=vmax)
  save_to_png(
      scaled,
      path=path,
      show=show,
      image_mode=image_mode,
      labels=labels,
      scale=scale)


def _deepvariant_channel_names(num_channels):
  """Get DeepVariant channel names for the given number of channels."""
  # Add additional empty labels if there are more channels than expected.
  filler_labels = [
      'channel {}'.format(i + 1)
      for i in range(len(DEEPVARIANT_CHANNEL_NAMES), num_channels)
  ]
  labels = DEEPVARIANT_CHANNEL_NAMES + filler_labels
  # Trim off any extra labels.
  return labels[0:num_channels]


def draw_deepvariant_pileup(example=None,
                            channels=None,
                            composite_type=None,
                            annotated=True,
                            labels=None,
                            path=None,
                            show=True,
                            scale=None):
  """Quick utility for showing a pileup example as channels or RGB.

  Args:
    example: A tensorflow Example containing image/encoded and image/shape
      features. Will be parsed through channels_from_example. Ignored if
      channels are provided directly. Either example OR channels is required.
    channels: list of 2D arrays containing the data to draw. Either example OR
      channels is required.
    composite_type: str or None. Method for combining channels. One of
      [None,"RGB"].
    annotated: bool. Whether to add channel labels and mark midpoints.
    labels: list of str. Which labels to add to the image. If annotated=True,
      use default channels labels for DeepVariant.
    path: str. Output file path for saving as an image. If None, just show plot.
    show: bool. Whether to display the image for ipython notebooks. Set to False
      to prevent extra output when running in bulk.
    scale: integer. Multiplier to enlarge the image. Default: None, which will
      set it automatically for a human-readable size. Set to 1 for no scaling.

  Returns:
    None. Saves an image at path and optionally shows it with IPython.display.
  """
  if example and not channels:
    channels = channels_from_example(example)
  elif not channels:
    raise ValueError('Either example OR channels must be specified.')

  if composite_type is None:
    img_array = np.concatenate(channels, axis=1)
    if annotated and labels is None:
      labels = _deepvariant_channel_names(len(channels))
  elif composite_type == 'RGB':
    img_array = convert_6_channels_to_rgb(channels)
    if annotated and labels is None:
      labels = ['']  # Creates one midpoint with no label.
  else:
    raise ValueError(
        "Unrecognized composite_type: {}. Must be None or 'RGB'".format(
            composite_type))

  array_to_png(
      img_array,
      path=path,
      show=show,
      scale=scale,
      labels=labels,
      vmin=0,
      vmax=254)


def variant_from_example(example):
  """Extract Variant object from the 'variant/encoded' feature of an Example.

  Args:
    example: a DeepVariant-style make_examples output example.

  Returns:
    A Nucleus Variant.
  """
  features = example.features.feature
  var_string = features['variant/encoded'].bytes_list.value[0]
  return variants_pb2.Variant.FromString(var_string)


def locus_id_from_variant(variant):
  """Create a locus ID of form "chr:pos_ref" from a Variant object.

  Args:
    variant: a nucleus variant.

  Returns:
    str.
  """
  return '{}:{}_{}'.format(variant.reference_name, variant.start,
                           variant.reference_bases)


def alt_allele_indices_from_example(example):
  """Extract indices of the particular alt allele(s) the example represents.

  Args:
    example: a DeepVariant make_examples output example.

  Returns:
    list of indices.
  """
  features = example.features.feature
  val = features['alt_allele_indices/encoded'].bytes_list.value[0]
  # Extract the encoded proto into unsigned integers and convert to regular ints
  mapped = [int(x) for x in np.frombuffer(val, dtype=np.uint8)]
  # Format is [<field id + type>, <number of elements in array>, ...<array>].
  # Extract the array only, leaving out the metadata.
  return mapped[2:]


def alt_bases_from_indices(alt_allele_indices, alternate_bases):
  """Get alt allele bases based on their indices.

  e.g. one alt allele: [0], ["C"] => "C"
  or with two alt alleles: [0,2], ["C", "TT", "A"] => "C-A"

  Args:
    alt_allele_indices: list of integers. Indices of the alt alleles for a
      particular example.
    alternate_bases: list of strings. All alternate alleles for the variant.

  Returns:
    str. Alt allele(s) at the indices, joined by '-' if more than 1.
  """
  alleles = [alternate_bases[i] for i in alt_allele_indices]
  # Avoiding '/' to support use in file paths.
  return '-'.join(alleles)


def alt_from_example(example):
  """Get alt allele(s) from a DeepVariant example.

  Args:
    example: a DeepVariant make_examples output example.

  Returns:
    str. The bases of the alt alleles, joined by a -.
  """
  variant = variant_from_example(example)
  indices = alt_allele_indices_from_example(example)
  return alt_bases_from_indices(indices, variant.alternate_bases)


def locus_id_with_alt(example):
  """Get complete locus ID from a DeepVariant example.

  Args:
    example: a DeepVariant make_examples output example.

  Returns:
    str in the form "chr:pos_ref_alt.
  """
  variant = variant_from_example(example)
  locus_id = locus_id_from_variant(variant)
  alt = alt_from_example(example)
  return '{}_{}'.format(locus_id, alt)


def label_from_example(example):
  """Get the "label" from an example.

  Args:
    example: a DeepVariant make_examples output example.

  Returns:
    integer (0, 1, or 2 for regular DeepVariant examples) or None if the
        example has no label.
  """
  val = example.features.feature['label'].int64_list.value
  if val:
    return int(val[0])
  else:
    return None



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Counts variants in a VCF, both by type and per chromosome."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import sys

from absl import app
from nucleus.io import vcf
from nucleus.util import variant_utils


def main(argv):
  if len(argv) != 2:
    print('Usage: {} <input_vcf>'.format(argv[0]))
    sys.exit(-1)
  in_vcf = argv[1]

  total = 0
  by_type = collections.defaultdict(int)
  by_ref = collections.defaultdict(int)

  with vcf.VcfReader(in_vcf) as reader:
    for variant in reader:
      total += 1
      by_type[variant_utils.variant_type(variant)] += 1
      by_ref[variant.reference_name] += 1

  print('# variants: {}'.format(total))
  print('# ref variants: {}'.format(by_type[variant_utils.VariantType.ref]))
  print('# SNP variants: {}'.format(by_type[variant_utils.VariantType.snp]))
  print('# indel variants: {}'.format(by_type[variant_utils.VariantType.indel]))
  for k, v in sorted(by_ref.items()):
    print('# variants in {}: {}'.format(k, v))


if __name__ == '__main__':
  app.run(main)



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Print an ASCII art pileup image."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl import app

from nucleus.io import sam
from nucleus.util import ranges

ANSI_BOLD = '\033[1m'
ANSI_RED = '\033[91m'
ANSI_OFF = '\033[0m'

# TODO(thomaswc): Also have this print out the reference sequence at the
# top if a reference fasta file is supplied.


def read_str(left_pos, start, highlight_position, seq):
  """Returns an aligned and highlighted ASCII representation of sequence."""
  s = ' ' * (start - left_pos)
  i = highlight_position - start
  j = i + 1
  s += seq[:i] + ANSI_BOLD + ANSI_RED + seq[i:j] + ANSI_OFF + seq[j:]
  return s


def ascii_pileup(sam_filename, query):
  """Returns an ASCII pileup image for the query as a list of strings.

  Args:
    sam_filename: The filename of the BAM/SAM file.
    query: String version of range.
  """
  r = ranges.parse_literal(query)
  position = r.start

  with sam.SamReader(sam_filename) as sam_reader:
    reads = sam_reader.query(r)
    pos_seq_pairs = sorted(
        (read.alignment.position.position, read.aligned_sequence)
        for read in reads)
    if not pos_seq_pairs:
      print('No overlapping reads found for', query)
      return []

    left_position = pos_seq_pairs[0][0]
    return [read_str(left_position, start, position, seq)
            for start, seq in pos_seq_pairs]


def main(argv):
  if len(argv) != 3:
    print('Usage: {} <input_sam> <chromosome>:<position>'.format(argv[0]))
    return -1
  in_sam = argv[1]
  query = argv[2]
  print('\n'.join(ascii_pileup(in_sam, query)))


if __name__ == '__main__':
  app.run(main)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: nucleus/protos/fastq.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='nucleus/protos/fastq.proto',
  package='nucleus.genomics.v1',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x1anucleus/protos/fastq.proto\x12\x13nucleus.genomics.v1\"Q\n\x0b\x46\x61stqRecord\x12\n\n\x02id\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12\x10\n\x08sequence\x18\x03 \x01(\t\x12\x0f\n\x07quality\x18\x04 \x01(\t\"8\n\x12\x46\x61stqReaderOptions\x12\x1c\n\x14skip_invalid_records\x18\x02 \x01(\x08J\x04\x08\x01\x10\x02\"\x14\n\x12\x46\x61stqWriterOptionsb\x06proto3')
)




_FASTQRECORD = _descriptor.Descriptor(
  name='FastqRecord',
  full_name='nucleus.genomics.v1.FastqRecord',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='nucleus.genomics.v1.FastqRecord.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='description', full_name='nucleus.genomics.v1.FastqRecord.description', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sequence', full_name='nucleus.genomics.v1.FastqRecord.sequence', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='quality', full_name='nucleus.genomics.v1.FastqRecord.quality', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=51,
  serialized_end=132,
)


_FASTQREADEROPTIONS = _descriptor.Descriptor(
  name='FastqReaderOptions',
  full_name='nucleus.genomics.v1.FastqReaderOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='skip_invalid_records', full_name='nucleus.genomics.v1.FastqReaderOptions.skip_invalid_records', index=0,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=134,
  serialized_end=190,
)


_FASTQWRITEROPTIONS = _descriptor.Descriptor(
  name='FastqWriterOptions',
  full_name='nucleus.genomics.v1.FastqWriterOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=192,
  serialized_end=212,
)

DESCRIPTOR.message_types_by_name['FastqRecord'] = _FASTQRECORD
DESCRIPTOR.message_types_by_name['FastqReaderOptions'] = _FASTQREADEROPTIONS
DESCRIPTOR.message_types_by_name['FastqWriterOptions'] = _FASTQWRITEROPTIONS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

FastqRecord = _reflection.GeneratedProtocolMessageType('FastqRecord', (_message.Message,), {
  'DESCRIPTOR' : _FASTQRECORD,
  '__module__' : 'nucleus.protos.fastq_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.FastqRecord)
  })
_sym_db.RegisterMessage(FastqRecord)

FastqReaderOptions = _reflection.GeneratedProtocolMessageType('FastqReaderOptions', (_message.Message,), {
  'DESCRIPTOR' : _FASTQREADEROPTIONS,
  '__module__' : 'nucleus.protos.fastq_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.FastqReaderOptions)
  })
_sym_db.RegisterMessage(FastqReaderOptions)

FastqWriterOptions = _reflection.GeneratedProtocolMessageType('FastqWriterOptions', (_message.Message,), {
  'DESCRIPTOR' : _FASTQWRITEROPTIONS,
  '__module__' : 'nucleus.protos.fastq_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.FastqWriterOptions)
  })
_sym_db.RegisterMessage(FastqWriterOptions)


# @@protoc_insertion_point(module_scope)



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Library of application-specific errors.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import contextlib
import errno
import sys

from absl import logging


class Error(Exception):
  """Base class for core error types."""


class CommandLineError(Error):
  """Exception class related to invalid command-line flags."""


def log_and_raise(msg, exception_class=Error):
  """Logs the given message at ERROR level and raises exception.

  Args:
    msg: [`string`]. The message to log and use in the raised exception.
    exception_class: [`Exception`]. The class of exception to raise.

  Raises:
    Error: An exception of the type specified by the input exception_class.
  """
  logging.error(msg)
  raise exception_class(msg)


@contextlib.contextmanager
def clean_commandline_error_exit(
    allowed_exceptions=(Error, CommandLineError), exit_value=errno.ENOENT):
  """Wraps commands to capture certain exceptions and exit without stacktraces.

  This function is intended to wrap all code within main() of Python binaries
  to provide a mechanism for user errors to exit abnormally without causing
  exceptions to be thrown. Any exceptions that are subclasses of those listed
  in `allowed_exceptions` will be caught and the program will quietly exit with
  `exit_value`. Other exceptions are propagated normally.

  NOTE: This function should only be used as a context manager and its usage
  should be limited to main().

  Args:
    allowed_exceptions: [`tuple of Exception`]. A tuple of Exception classes
      that should not be raised, but instead quietly caused to exit the program.
    exit_value: [`int`]. The value to return upon program exit.

  Yields:
    The yield in this function is used to allow the block nested in the "with"
    statement to be executed.
  """
  try:
    yield
  except allowed_exceptions:
    sys.exit(exit_value)



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fake setup.py module for installing Nucleus.

Usually, setup.py is invoked twice:  first, to build the pip package
and second to install it.

This setup.py is only used for installation; build_pip_package.sh is
used to create the package.  We do it this way because we need our
package to include symbolic links, which normal setup.py doesn't
support.

For the same reason, this setup.py is not implemented using setuptools.
Instead, we directly implement the four commands run by pip install
(https://pip.pypa.io/en/stable/reference/pip_install/#id46):
  * setup.py egg_info [--egg-base XXX]
  * setup.py install --record XXX [--single-version-externally-managed]
          [--root XXX] [--compile|--no-compile] [--install-headers XXX]
  * setup.py bdist_wheel -d XXX
  * setup.py clean
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from distutils import dist
import distutils.command.install as dist_install
import glob
import os
import shutil
import sys

# Basename of the .egg-info directory.
_EGG_DIR_BASENAME = 'google_nucleus.egg-info'


def _unsafe_filename(filename):
  return any(c in filename for c in ' \t\n\r;&<>*')


def touch(fname):
  open(fname, 'w+').close()


def find_destination(is_user):
  """Returns the directory we are supposed to install into."""
  install_cmd = dist_install.install(dist.Distribution())
  install_cmd.finalize_options()
  if is_user:
    return install_cmd.install_usersite
  else:
    return install_cmd.install_platlib


def copy_egg_info(dest_dir):
  """Copies the .egg-info directory to the specified location.

  Args:
    dest_dir: str. The destination directory.

  Returns:
    0 on success, 1 on failure.
  """
  # Remove any trailing slash on the destination directory.
  dest_dir = dest_dir.rstrip('/')
  egg_srcs = glob.glob('google_nucleus-*-py*.egg-info')
  if not egg_srcs:
    print('Could not find source .egg-info directory')
    return 1
  egg_src = egg_srcs[0]
  print('Copying egg-info from ', egg_src, ' to ', dest_dir)
  shutil.copytree(egg_src, dest_dir)
  return 0


def main():
  if len(sys.argv) < 2:
    print('Missing command')
    sys.exit(1)

  cmd = sys.argv[1]
  args = sys.argv[2:]

  if cmd == 'egg_info':
    egg_dir = _EGG_DIR_BASENAME
    if len(args) > 1 and args[0] == '--egg-base':
      egg_dir = os.path.join(args[1], egg_dir)
    sys.exit(copy_egg_info(egg_dir))

  elif cmd == 'install':
    destination = find_destination('--user' in args)

    record_file = 'install-record.txt'
    if '--record' in args:
      i = args.index('--record')
      record_file = args[i+1]
      if _unsafe_filename(record_file):
        print('Invalid record_file name', record_file)
        sys.exit(1)

    print('Removing old protobuf files')
    os.system('rm -fR ' + destination + '/google/protobuf')

    print('Installing Nucleus to ' + destination
          + ' with record file at ' + record_file)
    os.system('cp -R -v google nucleus ' + destination
              + " | awk '{print substr($3,2,length($3)-2)}' > " + record_file)
    # Append the .egg-info directory (must include trailing '/') to the record.
    dest_egg_dir = os.path.join(destination, _EGG_DIR_BASENAME) + '/'
    os.system('echo ' + dest_egg_dir + ' >> ' + record_file)
    # End by copying the .egg-info directory to the destination.
    sys.exit(copy_egg_info(dest_egg_dir))

  elif cmd == 'bdist_wheel':
    print('This package does not support wheel creation.')
    sys.exit(1)

  elif cmd == 'clean':
    # Nothing to do
    sys.exit(0)

  print('Unknown command: ', cmd)
  sys.exit(1)


if __name__ == '__main__':
  main()



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for Range overlap detection."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import re

from absl import logging
import intervaltree
import six

from nucleus.io import gfile
from nucleus.io import bed
from nucleus.protos import position_pb2
from nucleus.protos import range_pb2

# Regular expressions for matching literal chr:start-stop strings.
_REGION_LITERAL_REGEXP = re.compile(r'^(\S+):([0-9,]+)-([0-9,]+)$')

# Regular expressions for matching literal chr:start strings.
_POSITION_LITERAL_REGEXP = re.compile(r'^(\S+):([0-9,]+)$')

# Logging frequency when building our rangeset objects, which can take some time
# to complete. Rather than just pausing for a few minutes, we provide an update
# logging message every _LOG_EVERY_N_RANGES_IN_RANGESET_INIT records added. See
# b/110987941 for more information.
_LOG_EVERY_N_RANGES_IN_RANGESET_INIT = 250000


class RangeSet(object):
  """Fast overlap detection of a genomic position against a database of Ranges.

  Enables O(log n) computation of whether a point chr:pos falls within one of a
  large number of genomic ranges.

  This class does not supports overlapping or adjacent intervals. Any such
  intervals will be automatically merged together in the constructor.

  This class is immutable. No methods should be added that directly modify the
  ranges held by the class.
  """

  def __init__(self, ranges=None, contigs=None, quiet=False):
    """Creates a RangeSet backed by ranges.

    Note that the Range objects in ranges are *not* stored directly here, so
    they can safely be modified after they are passed to this RangeSet.

    Args:
      ranges: list(nucleus.genomics.v1.Range) protos (or anything with
        reference_name, start, and end properties following the Range
        convention). If None, no ranges will be used, and overlaps() will always
        return False.
      contigs: list(nucleus.genomics.v1.ContigInfo) protos. Used to define the
        iteration order over contigs (i.e., by contig.pos_in_fasta).  If this
        list is not provided, the iteration order will be determined by the
        alphabetical order of the contig names.
      quiet: bool; defaults to False: If False, we will emit a logging message
        every _LOG_EVERY_N_RANGES_IN_RANGESET_INIT records processed while
        building this intervaltree. Set to True to stop all of the logging.

    Raises:
      ValueError: if any range's reference_name does not correspond to any
        contig in `contigs`.
    """
    if contigs is not None:
      self._contigs = contigs
      self._contig_map = contigs_dict(contigs)
      self._contig_sort_key_fn = (
          lambda name: self._contig_map[name].pos_in_fasta)
      self._is_valid_contig = lambda name: name in self._contig_map
    else:
      self._contigs = None
      self._contig_map = None
      self._contig_sort_key_fn = lambda name: name
      self._is_valid_contig = lambda name: True

    if ranges is None:
      ranges = []

    # Add each range to our contig-specific intervaltrees.
    self._by_chr = collections.defaultdict(intervaltree.IntervalTree)
    for i, range_ in enumerate(ranges):
      if not self._is_valid_contig(range_.reference_name):
        raise ValueError(
            'Range {} is on an unrecognized contig.'.format(range_))
      self._by_chr[range_.reference_name].addi(range_.start, range_.end, None)
      if not quiet and i > 0 and i % _LOG_EVERY_N_RANGES_IN_RANGESET_INIT == 0:
        # We do our test directly here on i > 0 so we only see the log messages
        # if we add at least _LOG_EVERY_N_RANGES_IN_RANGESET_INIT records.
        logging.info('Adding interval %s to intervaltree', to_literal(range_))

    # Merge overlapping / adjacent intervals in each tree.
    for tree in six.itervalues(self._by_chr):
      tree.merge_overlaps(strict=False)

  def __iter__(self):
    """Iterate over the ranges in this RangeSet.

    Yields:
      Each range of this RangeSet, in sorted order (by chromosome, then start
      end positions). Relative ordering of chromosomes is defined by the
      contig.pos_in_fasta integer key for the associated contig. These objects
      are new range protos so can be freely modified.
    """
    for refname in sorted(
        six.iterkeys(self._by_chr), key=self._contig_sort_key_fn):
      for start, end, _ in sorted(self._by_chr[refname]):
        yield make_range(refname, start, end)

  @classmethod
  def from_regions(cls, regions, contig_map=None):
    """Parses a command-line style literal regions flag into a RangeSet.

    Args:
      regions: An iterable or None. If not None, regions will be parsed with
        ranges.from_regions.
      contig_map: An optional dictionary mapping from contig names to ContigInfo
        protobufs. If provided, allows literals of the format "contig_name",
        which will be parsed into a Range with reference_name=contig_name,
        start=0, end=n_bases where n_bases comes from the ContigInfo;
        additionally the sort order of the RangeSet will be determined by
        contig.pos_in_fasta.

    Returns:
      A RangeSet object.
    """
    if regions is None:
      return cls(ranges=[])
    else:
      return cls(ranges=from_regions(regions, contig_map=contig_map))

  @classmethod
  def from_contigs(cls, contigs):
    """Creates a RangeSet with an interval covering each base of each contig."""
    return cls(
        (make_range(contig.name, 0, contig.n_bases) for contig in contigs),
        contigs)

  @classmethod
  def from_bed(cls, source, contigs=None):
    """Creates a RangeSet containing the intervals from source.

    Args:
      source: A path to a BED (or equivalent) file of intervals.
      contigs: An optional list of ContigInfo proto, used by RangeSet
        constructor.

    Returns:
      A RangeSet.
    """
    return cls(bed_parser(source), contigs)

  def intersection(self, *others):
    """Computes the intersection among this RangeSet and *others RangeSets.

    This function computes the intersection of all of the intervals in self and
    *others, returning a RangeSet containing only intervals common to all. The
    intersection here is an ranged intersection, not an identity intersection,
    so the resulting set of intervals may not contain any of the original
    intervals in any of the sets.

    To be concrete, suppose we have three sets to intersect, each having two
    intervals:

      self   : chr1:1-10, chr2:20-30
      other1 : chr1:5-8, chr3:10-40
      other2 : chr1:3-7, chr3:10-30

    self.intersection(other1, other2) produces a RangeSet with one interval
    chr1:5-7, the common bases on chr1 in self, other1, and other2. No intervals
    on chr2 or chr3 are included since the chr2 only occurs in self and the two
    intervals on chr3, despite having some shared bases, don't have an
    overlapping interval in self.

    Args:
      *others: A list of RangeSet objects to intersect with the intervals in
        this RangeSet.

    Returns:
      A RangeSet. If *others is empty, this function returns self rather than
      making an unnecessary copy. In all other cases, the returned value will be
      a freshly allocated RangeSet.
    """

    def _intersect2(refname, tree1, tree2):
      """Intersects the intervals of two IntervalTrees."""
      # Yields all of the overlapping intervals from each interval of tree1
      # found in tree2. Since each tree has only non-adjacent, non-overlapping,
      # intervals this calculation is straightforward and safe and produces only
      # non-adjacent, non-overlapping intervals.
      if len(tree1) > len(tree2):
        (bigtree, smalltree) = (tree1, tree2)
      else:
        (bigtree, smalltree) = (tree2, tree1)
      return (make_range(refname, max(interval1.begin, overlapping.begin),
                         min(interval1.end, overlapping.end))
              for interval1 in bigtree
              for overlapping in smalltree.overlap(interval1))

    # Iteratively intersect each of our *other RangeSets with this RangeSet.
    # Sort by size so we do the smallest number of element merge first.
    # TODO(mdepristo): Note we could optimize this code by computing the set of
    # common contigs upfront across all others and only looping over those.
    intersected = self
    for other in sorted(others, key=len):
      intersected_intervals = []
      # pylint: disable=protected-access
      # So we can intersect intervals within each contig separately.
      for refname, intervals in six.iteritems(intersected._by_chr):
        # If refname is present in other, intersect those two IntervalTrees
        # directly and add those contigs to our growing list of intersected
        # intervals. If refname isn't present, all of the intervals on refname
        # should be dropped as there are no intervals to overlap.
        other_chr = other._by_chr.get(refname, None)
        if other_chr:
          intersected_intervals.extend(
              _intersect2(refname, intervals, other_chr))

      # Update our intersected RangeSet with the new intervals.
      intersected = RangeSet(intersected_intervals, self._contigs)

    return intersected

  def exclude_regions(self, other):
    """Chops out all of the intervals in other from this this RangeSet.

    NOTE: This is a *MUTATING* operation for performance reasons. Make a copy
    of self if you want to avoid modifying the RangeSet.

    Args:
      other: A RangeSet object whose intervals will be removed from this
        RangeSet.
    """
    # pylint: disable=protected-access
    for chrname, chr_intervals in six.iteritems(other._by_chr):
      # If refname is present in self, difference those two IntervalTrees.
      self_intervals = self._by_chr.get(chrname, None)
      if self_intervals:
        for begin, end, _ in chr_intervals:
          self_intervals.chop(begin, end)
        if self_intervals.is_empty():
          # Cleanup after ourselves by removing empty trees from our map.
          del self._by_chr[chrname]

  def __len__(self):
    """Gets the number of ranges used by this RangeSet."""
    return sum(len(for_chr) for for_chr in six.itervalues(self._by_chr))

  def __nonzero__(self):
    """Returns True if this RangeSet is not empty."""
    return bool(self._by_chr)

  __bool__ = __nonzero__  # Python 3 compatibility.

  def variant_overlaps(self, variant, empty_set_return_value=True):
    """Returns True if the variant's range overlaps with any in this set."""
    if not self:
      return empty_set_return_value
    else:
      return self.overlaps(variant.reference_name, variant.start)

  def overlaps(self, chrom, pos):
    """Returns True if chr:pos overlaps with any range in this RangeSet.

    Uses a fast bisection algorithm to determine the overlap in O(log n) time.

    Args:
      chrom: str. The chromosome name.
      pos: int. The position (0-based).

    Returns:
      True if chr:pos overlaps with a range.
    """
    chr_ranges = self._by_chr.get(chrom, None)
    if chr_ranges is None:
      return False
    return chr_ranges.overlaps(pos)

  def partition(self, max_size):
    """Splits our intervals so that none are larger than max_size.

    Slices up the intervals in this RangeSet into a equivalent set of intervals
    (i.e., spanning the same set of bases), each of which is at most max_size in
    length.

    This function does not modify this RangeSet.

    Because RangeSet merges adjacent intervals, this function cannot use a
    RangeSet to represent the partitioned intervals and so instead generates
    these intervals via a yield statement.

    Args:
      max_size: int > 0. The maximum size of any interval.

    Yields:
      nucleus.genomics.v1.Range protos, in sorted order (see comment about order
      in __iter__).

    Raises:
      ValueError: if max_size <= 0.
    """
    if max_size <= 0:
      raise ValueError('max_size must be > 0: {}'.format(max_size))

    for interval in self:
      refname = interval.reference_name
      for pos in range(interval.start, interval.end, max_size):
        yield make_range(refname, pos, min(interval.end, pos + max_size))

  def envelops(self, chrom, start, end):
    """Returns True iff some range in this RangeSet envelops the range.

    Args:
      chrom: str. The chromosome of interest.
      start: int. Zero-based inclusive index of the query range.
      end: int: Zero-based exclusive index of the query range.

    Returns:
      True if and only if some range in `self` completely spans the query
      range.
    """
    chr_ranges = self._by_chr.get(chrom, None)
    if chr_ranges is None:
      return False
    # The intervaltree package does the inverse check, i.e. whether ranges
    # contained in it overlap with the query region. So it returns nothing
    # when start == end. We by convention want anything overlapping the start
    # position to still indicate enveloping in this case.
    if start == end:
      return chr_ranges.overlaps(start)
    else:
      overlap_set = chr_ranges.overlap(begin=start, end=end)
      return any(ov.begin <= start and ov.end >= end for ov in overlap_set)


def make_position(chrom, position, reverse_strand=False):
  """Returns a nucleus.genomics.v1.Position.

  Args:
    chrom: str. The chromosome name.
    position: int. The start position (0-based, inclusive).
    reverse_strand: bool. If True, indicates the position is on the negative
      strand.
  """
  return position_pb2.Position(
      reference_name=chrom, position=position, reverse_strand=reverse_strand)


def make_range(chrom, start, end):
  """Returns a nucleus.genomics.v1.Range.

  Args:
    chrom: str. The chromosome name.
    start: int. The start position (0-based, inclusive) of this range.
    end: int. The end position (0-based, exclusive) of this range.

  Returns:
    A nucleus.genomics.v1.Range.
  """
  return range_pb2.Range(reference_name=chrom, start=start, end=end)


def position_overlaps(chrom, pos, interval):
  """Returns True iff the position chr:pos overlaps the interval.

  Args:
    chrom: str. The chromosome name.
    pos: int. The position (0-based, inclusive).
    interval: nucleus.genomics.v1.Range object.

  Returns:
    True if interval overlaps chr:pos.
  """
  return (chrom == interval.reference_name and
          interval.start <= pos < interval.end)


def ranges_overlap(i1, i2):
  """Returns True iff ranges i1 and i2 overlap.

  Args:
    i1: nucleus.genomics.v1.Range object.
    i2: nucleus.genomics.v1.Range object.

  Returns:
    True if and only if i1 and i2 overlap.
  """
  return (i1.reference_name == i2.reference_name and i1.end > i2.start and
          i1.start < i2.end)


def bedpe_parser(filename):
  """Parses Range objects from a BEDPE-formatted file object.

  See http://bedtools.readthedocs.org/en/latest/content/general-usage.html
  for more information on the BEDPE format.

  Skips events that span across chromosomes. For example, if the starting
  location is on chr1 and the ending location is on chr2, that record will
  not appear in the output.

  Args:
    filename: file name of a BEDPE-formatted file.

  Yields:
    nucleus.genomics.v1.Range protobuf objects.
  """
  for line in gfile.Open(filename):
    parts = line.split('\t')
    if parts[0] == parts[3]:
      # only keep events on the same chromosome
      yield make_range(parts[0], int(parts[1]), int(parts[5]))


def bed_parser(filename):
  """Parses Range objects from a BED-formatted file object.

  See http://bedtools.readthedocs.org/en/latest/content/general-usage.html
  for more information on the BED format.

  Args:
    filename: file name of a BED-formatted file.

  Yields:
    nucleus.genomics.v1.Range protobuf objects.
  """
  with bed.BedReader(filename) as fin:
    for r in fin.iterate():
      yield make_range(r.reference_name, r.start, r.end)


def from_regions(regions, contig_map=None):
  """Parses each region of `regions` into a Range proto.

  This function provides a super high-level interface for
  reading/parsing/converting objects into Range protos. Each `region` of
  `regions` is processed in turn, yielding one or more Range protos. This
  function inspects the contents of `region` to determine how to convert it to
  Range(s) protos. The following types of `region` strings are supported:

    * If region ends with an extension known in _get_parser_for_file, we treat
      region as a file and read the Range protos from it with the corresponding
      reader from _get_parser_for_file, yielding each Range from the file in
      order.
    * Otherwise we parse region as a region literal (`chr20:1-10`) and return
      the Range proto.

  Args:
    regions: iterable[str]. Converts each element of this iterable into
      region(s).
    contig_map: An optional dictionary mapping from contig names to ContigInfo
      protobufs. If provided, allows literals of the format "contig_name",
      which will be parsed into a Range with reference_name=contig_name,
      start=0, end=n_bases where n_bases comes from the ContigInfo.

  Yields:
    A Range proto.
  """
  for region in regions:
    reader = _get_parser_for_file(region)
    if reader:
      for elt in reader(region):
        yield elt
    else:
      yield parse_literal(region, contig_map)


# Cannot be at the top of the file because these parser functions need to be
# defined before adding them to the dictionary.
_REGION_FILE_READERS = {
    bed_parser: frozenset(['.bed']),
    bedpe_parser: frozenset(['.bedpe']),
}


def _get_parser_for_file(filename):
  for reader, exts in six.iteritems(_REGION_FILE_READERS):
    if any(filename.lower().endswith(ext) for ext in exts):
      return reader
  return None


def to_literal(range_pb):
  """Converts Range protobuf into string literal form.

  The string literal form looks like:

    reference_name:start+1-end

  since start and end are zero-based inclusive (start) and exclusive (end),
  while the literal form is one-based inclusive on both ends.

  Args:
    range_pb: A nucleus.genomics.v1.Range object.

  Returns:
    A string representation of the Range.
  """
  return '{}:{}-{}'.format(range_pb.reference_name, range_pb.start + 1,
                           range_pb.end)


def parse_literal(region_literal, contig_map=None):
  """Parses a Range from a string representation like chr:start-end.

  The region literal must conform to the following pattern:

    chromosome:start-end
    chromosome:position
    chromosome  [if contig_map is provided]

  chromosome can be any non-empty string without whitespace. start and end must
  both be positive integers. They can contain commas for readability. start and
  end are positions not offsets, so start == 1 means an offset of zero. If only
  a single position is provided, this creates a 1 bp interval starting at
  position - 1 and ending at position.

  Inspired by the samtools region specification:
  http://www.htslib.org/doc/samtools.html

  Args:
    region_literal: str. The literal to parse.
    contig_map: An optional dictionary mapping from contig names to ContigInfo
      protobufs. If provided, allows literals of the format "contig_name", which
      will be parsed into a Range with reference_name=contig_name, start=0,
      end=n_bases where n_bases comes from the ContigInfo.

  Returns:
    nucleus.genomics.v1.Range.

  Raises:
    ValueError: if region_literal cannot be parsed.
  """

  def parse_position(pos_str):
    return int(pos_str.replace(',', ''))

  matched = _REGION_LITERAL_REGEXP.match(region_literal)
  if matched:
    chrom, start, end = matched.groups()
    return make_range(chrom, parse_position(start) - 1, parse_position(end))

  matched = _POSITION_LITERAL_REGEXP.match(region_literal)
  if matched:
    chrom, pos = matched.groups()
    pos = parse_position(pos)
    return make_range(chrom, pos - 1, pos)

  if contig_map and region_literal in contig_map:
    # If the region_literals is an exact contig name like chr1 or MT return a
    # range over the entire contig.
    return make_range(region_literal, 0, contig_map[region_literal].n_bases)
  raise ValueError(
      'Could not parse "{}" as a region literal.  Region literals '
      'should have the form "chr:start-stop" or "chr:start" or '
      'just "chr".  A common error is to use the "chr" prefix on '
      'inputs that don\'t have it, or vice-versa.'.format(region_literal))


def parse_literals(region_literals, contig_map=None):
  """Parses each literal of region_literals in order."""
  return [parse_literal(literal, contig_map) for literal in region_literals]


def contigs_n_bases(contigs):
  """Returns the sum of all n_bases of contigs."""
  return sum(c.n_bases for c in contigs)


def contigs_dict(contigs):
  """Creates a dictionary for contigs.

  Args:
    contigs: Iterable of ContigInfo protos.

  Returns:
    A dictionary mapping contig.name: contig for each contig in contigs.
  """
  return {contig.name: contig for contig in contigs}


def sorted_ranges(ranges, contigs=None):
  """Sorts ranges by reference_name, start, and end.

  Args:
    ranges: Iterable of nucleus.genomics.v1.Range protos that we want to sort.
    contigs: None or an iterable of ContigInfo protos. If not None, we will use
      the order of the contigs (as defined by their pos_in_fasta field values)
      to sort the Ranges on different contigs with respect to each other.

  Returns:
    A newly allocated list of nucleus.genomics.v1.Range protos.
  """
  if contigs:
    contig_map = contigs_dict(contigs)

    def to_key(range_):
      pos = contig_map[range_.reference_name].pos_in_fasta
      return pos, range_.start, range_.end
  else:
    to_key = as_tuple

  return sorted(ranges, key=to_key)


def as_tuple(range_):
  """Returns a Python tuple (reference_name, start, end)."""
  return range_.reference_name, range_.start, range_.end


def overlap_len(range1, range2):
  """Computes the number of overlapping bases of range1 and range2.

  Args:
    range1: nucleus.genomics.v1.Range.
    range2: nucleus.genomics.v1.Range.

  Returns:
    int. The number of basepairs in common. 0 if the ranges are not on the same
    contig.
  """
  if range1.reference_name != range2.reference_name:
    return 0
  return max(0, (min(range1.end, range2.end) - max(range1.start, range2.start)))


def find_max_overlapping(query_range, search_ranges):
  """Gets the index of the element in search_ranges with max overlap with query.

  In case of ties, selects the lowest index range in search_ranges.

  Args:
    query_range: nucleus.genomics.v1.Range, read genomic range.
    search_ranges: list[nucleus.genomics.v1.Read]. The list of regions we want
      to search for the maximal overlap with query_range. NOTE: this must be a
      list (not a generator) as we loop over the search_ranges multiple times.

  Returns:
    int, the search_ranges index with the maximum read overlap. Returns None
    when read has no overlap with any of the search_ranges or search_ranges is
    empty.
  """
  if not search_ranges:
    return None
  overlaps = [overlap_len(query_range, srange) for srange in search_ranges]
  argmax = max(range(len(search_ranges)), key=lambda i: overlaps[i])
  # We return None if the read doesn't overlap at all.
  return None if overlaps[argmax] == 0 else argmax


def expand(region, n_bp, contig_map=None):
  """Expands region by n_bp in both directions.

  Takes a Range(chrom, start, stop) and returns a new
  Range(chrom, new_start, new_stop), where:

  -- new_start is max(start - n_bp, 0)
  -- new_stop is stop + n_bp if contig_map is None, or min(stop + n_bp, max_bp)
     where max_bp is contig_map[chrom].n_bp.

  Args:
    region: A nucleus.genomics.v1.Range proto.
    n_bp: int >= 0. how many basepairs to increase region by.
    contig_map: dict[string, ContigInfo] or None. If not None, used to get the
      maximum extent to increase stop by. Must have region.reference_name as a
      key.

  Returns:
    nucleus.genomics.v1.Range proto.

  Raises:
    ValueError: if n_bp is invalid.
    KeyError: contig_map is not None and region.reference_name isn't a key.
  """
  if n_bp < 0:
    raise ValueError('n_bp must be >= 0 but got {}'.format(n_bp))

  new_start = max(region.start - n_bp, 0)
  new_end = region.end + n_bp
  if contig_map is not None:
    new_end = min(new_end, contig_map[region.reference_name].n_bases)
  return make_range(region.reference_name, new_start, new_end)


def span(regions):
  """Returns a region that spans all of the bases in regions.

  This function returns a Range(chrom, start, stop), where start is the min
  of the starts in regions, and stop is the max end in regions. It may not be
  freshly allocated.

  Args:
    regions: list[Range]: a list of Range protos.

  Returns:
    A single Range proto.

  Raises:
    ValueError: if not all regions have the same reference_name.
    ValueError: if regions is empty.
  """
  if not regions:
    raise ValueError('regions is empty but must have at least one region')
  elif len(regions) == 1:
    return regions[0]
  elif any(r.reference_name != regions[0].reference_name for r in regions):
    raise ValueError('regions must be all on the same contig')
  else:
    start = min(r.start for r in regions)
    end = max(r.end for r in regions)
    return make_range(regions[0].reference_name, start, end)


def length(region):
  """Returns the length in basepairs of region."""
  return region.end - region.start



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Mathematics functions for working with genomics data.

A quick note on terminology here.

There are a bunch kinds of probabilities used commonly in genomics:

-- pError: the probability of being wrong.
-- pTrue: the probability of being correct.

Normalized probabilities vs. unnormalized likelihoods:

-- Normalized probabilities: p_1, ..., p_n such that sum(p_i) == 1 are said
   said to be normalized because they represent a valid probability
   distribution over the states 1 ... n.
-- Unnormalized likelihoods: p_1, ..., p_n where sum(p_i) != 1. These are not
   normalized and so aren't a valid probabilities distribution.

To add even more complexity, probabilities are often represented in three
semi-equivalent spaces:

-- Real-space: the classic space, with values ranging from [0.0, 1.0]
   inclusive.
-- log10-space: If p is the real-space value, in log10-space this would be
   represented as log10(p). How the p == 0 case is handled is often function
   dependent, which may accept/return -Inf or not handle the case entirely.
-- Phred-scaled: See https://en.wikipedia.org/wiki/Phred_quality_score for
   more information. Briefly, the Phred-scale maintains resolution in the lower
   parts of the probability space using integer quality scores (though using
   ints is optional, really). The phred-scale is defined as

     `phred(p) = -10 * log10(p)`

   where p is a real-space probability.

The functions in math.h dealing with probabilities are very explicit about what
kinds of probability and representation they expect and return, as unfortunately
these are all commonly represented as doubles in C++. Though it is tempting to
address this issue with classic software engineering practices like creating
a Probability class, in practice this is extremely difficult to do as this
code is often performance critical and the low-level mathematical operations
used in this code (e.g., log10) don't distiguish themselves among the types
of probabilities.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import numpy as np

from nucleus.util.python import math as math_

# C++ CLIF functions:
#
# We are enumerating the C++ functions exported by python/math.clif here, so
# it's clear to people what functions are available in python without digging
# into the raw python/C++ CLIF code.
log10_ptrue_to_phred = math_.log10_ptrue_to_phred
phred_to_perror = math_.phred_to_perror
phred_to_log10_perror = math_.phred_to_log10_perror
perror_to_log10_perror = math_.perror_to_log10_perror
perror_to_phred = math_.perror_to_phred
log10_perror_to_phred = math_.log10_perror_to_phred
perror_to_rounded_phred = math_.perror_to_rounded_phred
log10_perror_to_rounded_phred = math_.log10_perror_to_rounded_phred
log10_perror_to_perror = math_.log10_perror_to_perror
zero_shift_log10_probs = math_.zero_shift_log10_probs

# Maximum confidence in a variant call. Used to prevent overflow with log10.
# Note: -10 * log_10(1.25e-10) ~= 99.
_MAX_CONFIDENCE = 1.0 - 1.25e-10

LOG_10_OF_E = np.log10(np.e)
LOG_E_OF_10 = np.log(10.0)


def perror_to_bounded_log10_perror(perror, min_prob=1.0 - _MAX_CONFIDENCE):
  """Computes log10(p) for the given probability.

  The log10 probability is capped by -_MAX_CONFIDENCE.

  Args:
    perror: float. The probability to log10.
    min_prob: float. The minimum allowed probability.

  Returns:
    log10(p).

  Raises:
    ValueError: If probability is outside of [0.0, 1.0].
  """
  if not 0 <= perror <= 1:
    raise ValueError('perror must be between zero and one: {}'.format(perror))
  return perror_to_log10_perror(max(perror, min_prob))


def ptrue_to_bounded_phred(ptrue, max_prob=_MAX_CONFIDENCE):
  """Computes the Phred-scaled confidence from the given ptrue probability.

  See https://en.wikipedia.org/wiki/Phred_quality_score for more information.
  The quality score is capped by _MAX_CONFIDENCE.

  Args:
    ptrue: float. The ptrue probability to Phred scale.
    max_prob: float. The maximum allowed probability.

  Returns:
    Phred-scaled version of 1 - ptrue.

  Raises:
    ValueError: If ptrue is outside of [0.0, 1.0].
  """
  if not 0 <= ptrue <= 1:
    raise ValueError('ptrue must be between zero and one: {}'.format(ptrue))
  return perror_to_phred(1.0 - min(ptrue, max_prob))


def log10_binomial(k, n, p):
  """Calculates numerically-stable value of log10(binomial(k, n, p)).

  Returns the log10 of the binomial density for k successes in n trials where
  each success has a probability of occurring of p.

  In real-space, we would calculate:

     result = (n choose k) * (1-p)^(n-k) * p^k

  This function computes the log10 of result, which is:

     log10(result) = log10(n choose k) + (n-k) * log10(1-p) + k * log10(p)

  This is equivalent to invoking the R function:
    dbinom(x=k, size=n, prob=p, log=TRUE)

  See https://stat.ethz.ch/R-manual/R-devel/library/stats/html/Binomial.html
  for more details on the binomial.

  Args:
    k: int >= 0. Number of successes.
    n: int >= k. Number of trials.
    p: 0.0 <= float <= 1.0. Probability of success.

  Returns:
    log10 probability of seeing k successes in n trials with p.
  """
  r = math.lgamma(n + 1) - (math.lgamma(k + 1) + math.lgamma(n - k + 1))
  if k > 0:
    r += k * math.log(p)
  if n > k:
    r += (n-k) * math.log1p(-p)
  return r / LOG_E_OF_10


def log10sumexp(log10_probs):
  """Returns log10(sum(10^log10_probs)) computed in a numerically-stable way.

  Args:
    log10_probs: array-like of floats. An array of log10 probabilties.

  Returns:
    Float.
  """
  m = max(log10_probs)
  return m + math.log10(sum(pow(10.0, x - m) for x in log10_probs))


def normalize_log10_probs(log10_probs):
  """Approximately normalizes log10 probabilities.

  This function normalizes log10 probabilities. What this means is that we
  return an equivalent array of probabilities but whereas sum(10^log10_probs) is
  not necessarily 1.0, the resulting array is sum(10^result) ~= 1.0. The ~=
  indicates that the result is not necessarily == 1.0 but very close.

  This function is a fast and robust approximation of the true normalization of
  a log10 transformed probability vector. To understand the approximation,
  let's start with the exact calculation. Suppose I have three models, each
  emitting a probability that some data was generated by that model:

    data = {0.1, 0.01, 0.001} => probabilities from models A, B, and C

  These probabilities are unnormalized, in the sense that the total probability
  over the vector doesn't sum to 1 (sum(data) = 0.111). In many applications we
  want to normalize this vector so that sum(normalized(data)) = 1 and the
  relative magnitudes of the original probabilities are preserved (i.e,:

    data[i] / data[j] = normalized(data)[i] / normalized(data)[j]

  for all pairs of values indexed by i and j. For much of the work we do in
  genomics, we have so much data that representing these raw probability
  vectors in real-space risks numeric underflow/overflow, so we instead
  represent our probability vectors in log10 space:

    log10_data = log10(data) = {-1, -2, -3}

  Given that we expect numeric problems in real-space, normalizing this log10
  vector is hard, because the standard way you'd do the normalization is via:

    data[i] = data[i] / sum(data)
    log10_data[i] = log10_data[i] - log10(sum(10^data))

  But computing the sum of log10 values this way is dangerous because the naive
  implementation converts back to real-space to do the sum, the very operation
  we're trying to avoid due to numeric instability.

  This function implements an approximate normalization, which relaxes the need
  for an exact calculation of the sum. This function ensures that the
  normalization is numerically safe at the expense of the sum not being exactly
  equal to 1 but rather just close.

  Args:
    log10_probs: array-like of floats. An array of log10 probabilties.

  Returns:
    np.array with the same shape as log10_probs but where sum(10^result) ~= 1.0.

  Raises:
    ValueError: if any log10_probs > 0.0
  """
  log10_probs = np.array(log10_probs)
  if np.max(log10_probs) > 0.0:
    raise ValueError('log10_probs all must be <= 0', log10_probs)
  lse = log10sumexp(log10_probs)
  # np.minimum protects us from producing values slightly > 0.0 (e.g., 1e-16).
  return np.minimum(log10_probs - lse, 0.0)



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A Python interface for files."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six

from nucleus.io.python import gfile


def Exists(filename):
  return gfile.Exists(filename)


def Glob(pattern):
  return gfile.Glob(pattern)


class ReadableFile(six.Iterator):
  """Wraps gfile.ReadableFile to add iteration, enter/exit and readlines."""

  def __init__(self, filename):
    self._file = gfile.ReadableFile.New(filename)

  def __enter__(self):
    return self

  def __exit__(self, type_, value, traceback):
    self._file.__exit__()

  def __iter__(self):
    return self

  def __next__(self):
    ok, line = self._file.Readline()
    if ok:
      return line
    else:
      raise StopIteration

  def readlines(self):
    lines = []
    while True:
      ok, line = self._file.Readline()
      if ok:
        lines.append(line)
      else:
        break
    return lines


def Open(filename, mode="r"):
  if mode.startswith("r"):
    return ReadableFile(filename)
  elif mode.startswith("w"):
    return gfile.WritableFile.New(filename)
  else:
    raise ValueError("Unsupported mode '{}' for Open".format(mode))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: nucleus/protos/cigar.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='nucleus/protos/cigar.proto',
  package='nucleus.genomics.v1',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x1anucleus/protos/cigar.proto\x12\x13nucleus.genomics.v1\"\xb0\x02\n\tCigarUnit\x12;\n\toperation\x18\x01 \x01(\x0e\x32(.nucleus.genomics.v1.CigarUnit.Operation\x12\x18\n\x10operation_length\x18\x02 \x01(\x03\x12\x1a\n\x12reference_sequence\x18\x03 \x01(\t\"\xaf\x01\n\tOperation\x12\x19\n\x15OPERATION_UNSPECIFIED\x10\x00\x12\x13\n\x0f\x41LIGNMENT_MATCH\x10\x01\x12\n\n\x06INSERT\x10\x02\x12\n\n\x06\x44\x45LETE\x10\x03\x12\x08\n\x04SKIP\x10\x04\x12\r\n\tCLIP_SOFT\x10\x05\x12\r\n\tCLIP_HARD\x10\x06\x12\x07\n\x03PAD\x10\x07\x12\x12\n\x0eSEQUENCE_MATCH\x10\x08\x12\x15\n\x11SEQUENCE_MISMATCH\x10\tb\x06proto3')
)



_CIGARUNIT_OPERATION = _descriptor.EnumDescriptor(
  name='Operation',
  full_name='nucleus.genomics.v1.CigarUnit.Operation',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='OPERATION_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ALIGNMENT_MATCH', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INSERT', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DELETE', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SKIP', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CLIP_SOFT', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CLIP_HARD', index=6, number=6,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PAD', index=7, number=7,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SEQUENCE_MATCH', index=8, number=8,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SEQUENCE_MISMATCH', index=9, number=9,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=181,
  serialized_end=356,
)
_sym_db.RegisterEnumDescriptor(_CIGARUNIT_OPERATION)


_CIGARUNIT = _descriptor.Descriptor(
  name='CigarUnit',
  full_name='nucleus.genomics.v1.CigarUnit',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='operation', full_name='nucleus.genomics.v1.CigarUnit.operation', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='operation_length', full_name='nucleus.genomics.v1.CigarUnit.operation_length', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='reference_sequence', full_name='nucleus.genomics.v1.CigarUnit.reference_sequence', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _CIGARUNIT_OPERATION,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=52,
  serialized_end=356,
)

_CIGARUNIT.fields_by_name['operation'].enum_type = _CIGARUNIT_OPERATION
_CIGARUNIT_OPERATION.containing_type = _CIGARUNIT
DESCRIPTOR.message_types_by_name['CigarUnit'] = _CIGARUNIT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

CigarUnit = _reflection.GeneratedProtocolMessageType('CigarUnit', (_message.Message,), {
  'DESCRIPTOR' : _CIGARUNIT,
  '__module__' : 'nucleus.protos.cigar_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.CigarUnit)
  })
_sym_db.RegisterMessage(CigarUnit)


# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: nucleus/protos/reference.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from nucleus.protos import range_pb2 as nucleus_dot_protos_dot_range__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='nucleus/protos/reference.proto',
  package='nucleus.genomics.v1',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x1enucleus/protos/reference.proto\x12\x13nucleus.genomics.v1\x1a\x1anucleus/protos/range.proto\"\xbf\x01\n\nContigInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12\x0f\n\x07n_bases\x18\x03 \x01(\x03\x12\x39\n\x05\x65xtra\x18\x05 \x03(\x0b\x32*.nucleus.genomics.v1.ContigInfo.ExtraEntry\x12\x14\n\x0cpos_in_fasta\x18\x04 \x01(\x05\x1a,\n\nExtraEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"N\n\x11ReferenceSequence\x12*\n\x06region\x18\x01 \x01(\x0b\x32\x1a.nucleus.genomics.v1.Range\x12\r\n\x05\x62\x61ses\x18\x02 \x01(\tb\x06proto3')
  ,
  dependencies=[nucleus_dot_protos_dot_range__pb2.DESCRIPTOR,])




_CONTIGINFO_EXTRAENTRY = _descriptor.Descriptor(
  name='ExtraEntry',
  full_name='nucleus.genomics.v1.ContigInfo.ExtraEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='nucleus.genomics.v1.ContigInfo.ExtraEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='nucleus.genomics.v1.ContigInfo.ExtraEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=231,
  serialized_end=275,
)

_CONTIGINFO = _descriptor.Descriptor(
  name='ContigInfo',
  full_name='nucleus.genomics.v1.ContigInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='nucleus.genomics.v1.ContigInfo.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='description', full_name='nucleus.genomics.v1.ContigInfo.description', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='n_bases', full_name='nucleus.genomics.v1.ContigInfo.n_bases', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='extra', full_name='nucleus.genomics.v1.ContigInfo.extra', index=3,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pos_in_fasta', full_name='nucleus.genomics.v1.ContigInfo.pos_in_fasta', index=4,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_CONTIGINFO_EXTRAENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=84,
  serialized_end=275,
)


_REFERENCESEQUENCE = _descriptor.Descriptor(
  name='ReferenceSequence',
  full_name='nucleus.genomics.v1.ReferenceSequence',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='region', full_name='nucleus.genomics.v1.ReferenceSequence.region', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bases', full_name='nucleus.genomics.v1.ReferenceSequence.bases', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=277,
  serialized_end=355,
)

_CONTIGINFO_EXTRAENTRY.containing_type = _CONTIGINFO
_CONTIGINFO.fields_by_name['extra'].message_type = _CONTIGINFO_EXTRAENTRY
_REFERENCESEQUENCE.fields_by_name['region'].message_type = nucleus_dot_protos_dot_range__pb2._RANGE
DESCRIPTOR.message_types_by_name['ContigInfo'] = _CONTIGINFO
DESCRIPTOR.message_types_by_name['ReferenceSequence'] = _REFERENCESEQUENCE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ContigInfo = _reflection.GeneratedProtocolMessageType('ContigInfo', (_message.Message,), {

  'ExtraEntry' : _reflection.GeneratedProtocolMessageType('ExtraEntry', (_message.Message,), {
    'DESCRIPTOR' : _CONTIGINFO_EXTRAENTRY,
    '__module__' : 'nucleus.protos.reference_pb2'
    # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.ContigInfo.ExtraEntry)
    })
  ,
  'DESCRIPTOR' : _CONTIGINFO,
  '__module__' : 'nucleus.protos.reference_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.ContigInfo)
  })
_sym_db.RegisterMessage(ContigInfo)
_sym_db.RegisterMessage(ContigInfo.ExtraEntry)

ReferenceSequence = _reflection.GeneratedProtocolMessageType('ReferenceSequence', (_message.Message,), {
  'DESCRIPTOR' : _REFERENCESEQUENCE,
  '__module__' : 'nucleus.protos.reference_pb2'
  # @@protoc_insertion_point(class_scope:nucleus.genomics.v1.ReferenceSequence)
  })
_sym_db.RegisterMessage(ReferenceSequence)


_CONTIGINFO_EXTRAENTRY._options = None
# @@protoc_insertion_point(module_scope)



# Copyright 2018 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This example program adds the AD info field to a VCF file.

It assumes that the AD field of the individual variant calls is already
populated.

Sample usage:
  $ add_ad_to_vcf input.vcf.gz output.vcf.gz
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys

from absl import app
import six
from nucleus.io import vcf
from nucleus.util import variant_utils
from nucleus.util import variantcall_utils
from nucleus.util import vcf_constants


def get_variant_ad(variant):
  """Returns the allele depth for the Variant, calculated across its calls."""
  num_alleles = len(variant.alternate_bases) + 1
  call_ads = [variantcall_utils.get_format(vc, 'AD') for vc in
              variant.calls]
  assert(len(call_ad) == num_alleles for call_ad in call_ads)
  return [sum(call_ad[i] for call_ad in call_ads)
          for i in six.moves.xrange(num_alleles)]


def main(argv):
  if len(argv) != 3:
    print('Usage: %s <input_vcf> <output_vcf>' % argv[0])
    sys.exit(-1)
  in_vcf = argv[1]
  out_vcf = argv[2]

  with vcf.VcfReader(in_vcf) as reader:
    if 'AD' in [info.id for info in reader.header.infos]:
      print('%s already contains AD field.' % in_vcf)
      sys.exit(-1)
    out_header = reader.header
    out_header.infos.extend([vcf_constants.reserved_info_field('AD')])

    with vcf.VcfWriter(out_vcf, header=out_header) as writer:
      for variant in reader:
        variant_utils.set_info(variant, 'AD', get_variant_ad(variant),
                               writer)
        writer.write(variant)


if __name__ == '__main__':
  app.run(main)
